(* $Id: da.ml,v 1.56 2011-07-13 14:35:17 deraugla Exp $ *)

open Dim_def;
open Rt;
open Printf;

type da_data =
  { postscript : mutable option out_channel;
    text_buf : Buffer.t;
    ps_A3 : mutable bool;
    ps_landscape : mutable bool;
    rt_dim : mutable damdim;
    ps_dim : mutable damdim;
    sched_width : mutable int;
    sched_height : mutable int;
    inter_task_dy : mutable int;
    da_co : common }
and common =
  { xd : Rt.xdata;
    xa : Rt.xargs;
    c_white_col : (Rt.color * (int * int * int));
    c_black_col : (Rt.color * (int * int * int));
    c_red_col : (Rt.color * (int * int * int));
    c_blue_col : (Rt.color * (int * int * int));
    c_backg_rep_col : (Rt.color * (int * int * int));
    c_gray_col : (Rt.color * (int * int * int));
    c_color : array (Rt.color * (int * int * int));
    c_raw_font : Rt.font;
    pixmap_sched : Rt.pixmap;
    pixmap_dameid : Rt.pixmap;
    pixmap_caption : Rt.pixmap;
    watch : Rt.mouse;
    hand : Rt.mouse }
;

type da_color =
  { da_color : color;
    da_rgb : (int * int * int);
    da_c_dd : da_data }
and da_pixmap =
  { da_pixmap : pixmap;
    da_p_name : string;
    da_p_dd : da_data }
and da_font =
  { da_font : font;
    da_f_size : int;
    da_f_name : string;
    da_f_dd : da_data }
;

type da_paint =
  [ DaColorPn of da_color
  | DaPixmapPn of da_pixmap
  | DaNonePn ]
;

type drawable = (da_data * string * Rt.widget);

(* Common interface rt/ps *)

open Dim;

value xcoord_of_coord (dd, _, _) (x, y) =
  let dim =
    match dd.postscript with
    [ Some _ -> dd.ps_dim
    | None -> dd.rt_dim ]
  in
  (x + dim.x_center, - y + dim.y_center)
;

value da_drawable_width (dd, win, wid) =
  match dd.postscript with
  [ Some _ ->
      match win with
      [ "sched" | "mesoid" -> 1218
      | "resp_f_time" -> 430
      | s -> failwith ("da_drawable_width " ^ win) ]
  | None ->
      if win = "sched" || win = "sched_zoom" then dd.sched_width
      else widget_width wid ]
;

value da_drawable_height (dd, win, wid) =
  match dd.postscript with
  [ Some _ ->
      match win with
      [ "resp_f_time" -> 390
      | _ -> failwith ("da_drawable_height " ^ win) ]
  | None ->
      widget_height wid ]
;

value da_drawable_dim (dd, win, wid) =
  match dd.postscript with
  [ Some _ -> dd.ps_dim
  | None -> dd.rt_dim ]
;

value da_set_line_width dd v =
  match dd.postscript with
  [ Some oc ->
      fprintf oc "[] 0 setdash %g setlinewidth\n" (max (float v) 0.5)
  | None ->
      rt_set_line_width dd.da_co.xd v ]
;

value da_set_dashes dd off list =
  match dd.postscript with
  [ Some oc ->
      let (s, _) =
        List.fold_left (fun (s, sep) x -> (sprintf "%s%d" sep x, " "))
          ("", "") list
      in
      fprintf oc "[%s] 0 setdash\n" s
  | None ->
      rt_set_dashes dd.da_co.xd off list ]
;

value da_create_color dd (col, rgb) =
  {da_color = col; da_rgb = rgb; da_c_dd = dd}
;
value da_black_color dd =
  let (col, rgb) = dd.da_co.c_black_col in
  {da_color = col; da_rgb = rgb; da_c_dd = dd}
;
value da_white_color dd =
  let (col, rgb) = dd.da_co.c_white_col in
  {da_color = col; da_rgb = rgb; da_c_dd = dd}
;

value da_create_pixmap dd name w h =
  {da_pixmap = rt_create_pixmap dd.da_co.xd w h; da_p_name = name;
   da_p_dd = dd}
;

value da_select_color color =
  match color.da_c_dd.postscript with
  [ Some oc ->
      let (r, g, b) = color.da_rgb in
      fprintf oc "%g %g %g setrgbcolor\n" (float r /. 256.) (float g /. 256.)
        (float b /. 256.)
  | None ->
      rt_select_color color.da_color ]
;

value da_select_pixmap pixmap =
  match pixmap.da_p_dd.postscript with
  [ Some oc -> fprintf oc "0 setgray 0 %s setpattern\n" pixmap.da_p_name
  | None -> rt_select_pixmap pixmap.da_pixmap ]
;

value da_select_paint =
  fun
  [ DaColorPn color -> da_select_color color
  | DaPixmapPn pixmap -> da_select_pixmap pixmap
  | DaNonePn -> () ]
;

value da_select_font font =
  match font.da_f_dd.postscript with
  [ Some oc ->
      let fn = "DaBold" in
      fprintf oc "/%s findfont [%d 0 0 %d 0 0] makefont setfont\n" fn
        font.da_f_size (-font.da_f_size)
  | None ->
      rt_select_font font.da_font ]
;

value ps_ang a = floor (float a /. 64. *. 10. +. 0.5) /. 10.;

value da_x win x =
  match win with
  [ "dameid" -> x + 40
  | "caption" -> x + 650
  | "sched" -> x + 40
(*
  | "mesoid" -> x + 40
*)
  | "resp_f_time" -> x + 50
  | s -> failwith ("da_x " ^ s) ]
;

value da_y win y =
  match win with
  [ "dameid" -> y + 400
  | "caption" -> y + 500
  | "sched" -> y + (*150*)50
(*
  | "mesoid" -> y + 50
*)
  | "resp_f_time" -> y - 150
  | s -> failwith ("da_y " ^ s) ]
;

value text_x_pos = 820;
value text_y_pos = 480;

value da_clear_widget (dd, win, wid) =
  match dd.postscript with
  [ Some oc -> ()
  | None ->
      match win with
      [ "caption" | "dameid" | "sched" -> do {
          let w = screen_width dd.da_co.xd in
          let h = screen_height dd.da_co.xd in
          rt_select_color (fst dd.da_co.c_white_col);
          let pixmap =
            if win = "caption" then dd.da_co.pixmap_caption
            else if win = "dameid" then dd.da_co.pixmap_dameid
            else dd.da_co.pixmap_sched
          in
          rt_fill_rectangle (PixmapDr pixmap) (0, 0, w, h)
        }
      | _ ->
          rt_clear_widget wid ] ]
;

value da_commit_display (dd, win, wid) =
  match dd.postscript with
  [ Some oc -> ()
  | None ->
      match win with
      [ "caption" | "dameid" | "sched" -> rt_clear_widget wid
      | _ -> () ] ]
;

value da_clear_area (gd, win, wid) (x, y, w, h) =
  match gd.postscript with
  [ Some oc -> do {
      fprintf oc "1 setgray\n";
      fprintf oc "%d %d moveto\n" (da_x win x) (da_y win y);
      fprintf oc "%d 0 rlineto 0 %d rlineto %d 0 rlineto closepath\n"
        (da_x win (x + w) - da_x win x) (da_y win (y + h) - da_y win y)
        (- da_x win (x + w) + da_x win x);
      fprintf oc "fill\n";
    }
  | None ->
      rt_clear_area wid (x, y, w, h) ]
;

value drawable_of_win dd win wid =
  match win with
  [ "caption" -> PixmapDr dd.da_co.pixmap_caption
  | "dameid" -> PixmapDr dd.da_co.pixmap_dameid
  | "sched" -> PixmapDr dd.da_co.pixmap_sched
  | _ -> WidgetDr wid ]
;

value da_draw_arc (dd, win, wid) (bb_x, bb_y, bb_w, bb_h, a, da) =
  match dd.postscript with
  [ Some oc -> do {
      let radius = bb_w / 2 in
      let xc = da_x win (bb_x + radius) in
      let yc = da_y win (bb_y + radius) in
      let a1 = ps_ang (- a - da) in
      let a2 = ps_ang (- a) in
      fprintf oc "%d %d %d %g %g arc\n" xc yc radius (min a1 a2) (max a1 a2);
      fprintf oc "stroke\n";
    }
  | None ->
      let drw = drawable_of_win dd win wid in
      rt_draw_arc drw (bb_x, bb_y, bb_w , bb_h, a, da) ]
;

value da_fill_arc (dd, win, wid) (bb_x, bb_y, bb_w, bb_h, a, da) =
  match dd.postscript with
  [ Some oc -> do {
      let radius = bb_w / 2 in
      let xc = da_x win (bb_x + radius) in
      let yc = da_y win (bb_y + radius) in
      fprintf oc "%d %d moveto\n" xc yc;
      let a1 = ps_ang (- a - da) in
      let a2 = ps_ang (- a) in
      fprintf oc "%d %d %d %g %g arc\n" xc yc radius (min a1 a2) (max a1 a2);
      fprintf oc "closepath fill\n";
    }
  | None ->
      let drw = drawable_of_win dd win wid in
      rt_fill_arc drw (bb_x, bb_y, bb_w, bb_h, a, da) ]
;

value da_fill_rectangle (dd, win, wid) (x, y, w, h) =
  match dd.postscript with
  [ Some oc -> do {
      fprintf oc "%d %d moveto\n" (da_x win x) (da_y win y);
      fprintf oc "%d 0 rlineto 0 %d rlineto %d 0 rlineto closepath\n" w h
        (-w);
      fprintf oc "fill\n";
    }
  | None ->
      let drw = drawable_of_win dd win wid in
      rt_fill_rectangle drw (x, y, w, h) ]
;

value da_draw_rectangle (dd, win, wid) (x, y, w, h) =
  match dd.postscript with
  [ Some oc -> do {
      fprintf oc "%d %d moveto\n" (da_x win x) (da_y win y);
      fprintf oc "%d 0 rlineto 0 %d rlineto %d 0 rlineto closepath\n" w h
        (-w);
      fprintf oc "stroke\n";
    }
  | None ->
      let drw = drawable_of_win dd win wid in
      rt_draw_rectangle drw (x, y, w, h) ]
;

value da_draw_line (dd, win, wid) (x1, y1) (x2, y2) =
  match dd.postscript with
  [ Some oc ->
      fprintf oc "%d %d moveto %d %d lineto stroke\n" (da_x win x1)
        (da_y win y1) (da_x win x2) (da_y win y2)
  | None ->
      let drw = drawable_of_win dd win wid in
      rt_draw_line drw (x1, y1) (x2, y2) ]
;

value da_draw_lines (dd, win, wid) xyl =
  match dd.postscript with
  [ Some oc ->
      match xyl with
      [ [(x, y) :: xyl] -> do {
          fprintf oc "%d %d moveto\n" (da_x win x) (da_y win y);
          List.iter
            (fun (x, y) ->
              fprintf oc "%d %d lineto\n" (da_x win x) (da_y win y))
            xyl;
          fprintf oc "stroke\n";
        }
      | [] -> () ]
  | None ->
      let drw = drawable_of_win dd win wid in
      rt_draw_lines drw xyl ]
;

value rev_implode l =
  let s = String.create (List.length l) in
  loop (String.length s - 1) l where rec loop i =
    fun
    [ [c :: l] -> do { String.unsafe_set s i c; loop (i - 1) l }
    | [] -> s ]
;

value balan_paren_string s =
  let balanced =
    loop 0 0 where rec loop lev i =
      if i = String.length s then lev = 0
      else
        match s.[i] with
        [ '(' -> loop (lev + 1) (i + 1)
        | ')' -> if lev > 0 then loop (lev - 1) (i + 1) else False
        | _ -> loop lev (i + 1) ]
  in
  if balanced then s
  else
    loop [] 0 where rec loop rev_cl i =
      if i = String.length s then rev_implode rev_cl
      else
        let rev_cl =
          match s.[i] with
          [ '(' | ')' -> ['\\' :: rev_cl]
          | _ -> rev_cl ]
        in
        loop [s.[i] :: rev_cl] (i + 1)
;

value da_erase_draw_string (dd, win, wid) (x, y) s =
  match dd.postscript with
  [ Some oc -> do {
      let s = Std_term.latin_1_of_utf_8 s in
      fprintf oc "%d %d moveto (%s) true charpath pathbbox\n" (da_x win x)
        (da_y win y) (balan_paren_string s);
      let d = 4 in
      fprintf oc "/ury exch %d add def /urx exch %d add def\n" d d;
      fprintf oc "/lly exch %d sub def /llx exch %d sub def\n" d d;
      fprintf oc "gsave newpath\n";
      fprintf oc
        "llx lly moveto llx ury lineto urx ury lineto urx lly lineto\n";
      fprintf oc "closepath 1 setgray fill grestore fill\n";
    }
  | None ->
      let drw = drawable_of_win dd win wid in
      rt_erase_draw_string drw (x, y) s ]
;

value da_draw_string (dd, win, wid) (x, y) s =
  match dd.postscript with
  [ Some oc ->
      let s = Std_term.latin_1_of_utf_8 s in
      fprintf oc "%d %d moveto (%s) show\n" (da_x win x)
        (da_y win y) (balan_paren_string s)
  | None ->
      let drw = drawable_of_win dd win wid in
      rt_draw_string drw (x, y) s ]
;

value da_term_send gd wid s =
  match gd.postscript with
  [ Some _ -> Buffer.add_string gd.text_buf s
  | None -> term_send wid s ]
;

value da_text_width font txt = rt_text_width font.da_font txt;

value da_font_size font =
  match font.da_f_dd.postscript with
  [ Some oc -> font.da_f_size
  | None -> rt_font_size font.da_font ]
;

value draw_ps_text oc s = do {
  let sz = 12 in
  fprintf oc "0 setgray\n";
  fprintf oc "/DaBold findfont [%d 0 0 %d 0 0] makefont setfont\n" sz (-sz);
  let flush_cl has_curr_pt x y rev_cl =
    if rev_cl = [] then has_curr_pt
    else do {
      if not has_curr_pt then fprintf oc "%d %d moveto " x y else ();
      fprintf oc "(%s) show\n" (rev_implode rev_cl);
      True
    }
  in
  let s = Std_term.latin_1_of_utf_8 s in
  loop False text_x_pos text_y_pos [] 0 where rec
  loop has_curr_pt x y rev_cl i =
    if i = String.length s then
      let _ : bool = flush_cl has_curr_pt x y rev_cl in
      ()
    else
      match s.[i] with
      [ '\r' -> loop has_curr_pt x y rev_cl (i + 1)
      | '\n' -> do {
          let _ : bool = flush_cl has_curr_pt x y rev_cl in
          loop False x (y + sz) [] (i + 1)
        }
      | '\027' -> do {
          let (rev_seq, i) =
            loop [] (i + 1) where rec loop rev_cl i =
              if i = String.length s then (rev_cl, i)
              else
                match s.[i] with
                [ 'a'..'z' | 'A'..'Z' -> ([s.[i] :: rev_cl], i + 1)
                | c -> loop [c :: rev_cl] (i + 1) ]
          in
          let has_curr_pt = flush_cl has_curr_pt x y rev_cl in
          match List.rev rev_seq with
          [ ['['; '0'; 'm'] ->
              fprintf oc "/DaNorm findfont [%d 0 0 %d 0 0] makefont setfont\n"
                sz (-sz)
          | ['['; '1'; 'm'] ->
              fprintf oc "/DaBold findfont [%d 0 0 %d 0 0] makefont setfont\n"
                sz (-sz)
          | _ ->
              match rev_implode rev_seq with
              [ "]256;0;0F" -> fprintf oc "1 0 0 setrgbcolor\n"
              | "]0F" | "]F" -> fprintf oc "0 setgray\n"
              | s -> () ] ];
          loop has_curr_pt x y [] i
        }
      | c -> loop has_curr_pt x y [c :: rev_cl] (i + 1) ]
};

value print_def_font_with_accents oc =
  fprintf oc "\
/deffont {
  findfont dup length dict
    exch {
      1 index /FID ne
      2 index /UniqueID ne and {
        2 index 3 1 roll put
      } { pop pop } ifelse
    } forall
    dup dup /Encoding get dup length array copy
    dup 8#307 /Ccedilla put
    dup 8#342 /acircumflex put
    dup 8#351 /eacute put
    dup 8#350 /egrave put
    dup 8#373 /ucircumflex put
    /Encoding exch put
  definefont
  pop
} def
/DaNorm /Courier deffont
/DaBold /Courier-Bold deffont\n"
;

value postscript_save expose dd oc = do {
  fprintf oc "%%!\n";
  let v1 = 2100 * 72 / 254 in
  let v2 = 2970 * 72 / 254 in
  let (llx, lly) =
    if dd.ps_landscape then (0, 0)
    else if dd.ps_A3 then (0, v1)
    else (0, v2 / 2)
  in
  let (urx, ury) = if dd.ps_A3 then (v2, 2 * v1) else (v1, v2) in
  fprintf oc "%%%%BoundingBox: %d %d %d %d\n" llx lly urx ury;
  if dd.ps_A3 then fprintf oc "2 sqrt dup scale\n" else ();
  fprintf oc "/cm { 2.54 div 72 mul } def\n";
  dd.ps_A3 := False;
  let scale = 0.65 in
  if dd.ps_landscape then do {
    fprintf oc "90 rotate\n";
    fprintf oc "%.2f %.2f scale\n" scale (-. scale);
  }
  else do {
    let scale = scale /. sqrt 2. in
    fprintf oc "0 29.7 cm translate\n";
    fprintf oc "%.2f %.2f scale\n" scale (-. scale);
  };
  print_def_font_with_accents oc;
  fprintf oc "\
/Tile8x8 {
  /PaintType 2 /PatternType 1 /TilingType 1 /BBox [0 0 8 8] /XStep 8 /YStep 8
} bind def
/pixm1 <<
  Tile8x8
  /PaintProc {
    pop 0.5 setlinewidth 0 0 moveto 8 8 lineto 0 8 moveto 8 0 lineto stroke
  }
>> matrix makepattern def
/pixm2 <<
  Tile8x8
  /PaintProc {
    pop 0.5 setlinewidth 0 0 moveto 8 8 lineto 0 8 moveto 8 0 lineto stroke
    0 4 moveto 4 8 lineto 8 4 lineto 4 0 lineto 0 4 lineto stroke
  }
>> matrix makepattern def
/pixm3 <<
  Tile8x8
  /PaintProc {
    pop 0.5 setlinewidth -4 8 moveto 8 -4 lineto 0 12 moveto 12 0 lineto
    stroke
  }
>> matrix makepattern def
/pixm4 <<
  Tile8x8
  /PaintProc {
    pop 0.5 setlinewidth -2 0 moveto 4 12 lineto 0 -4 moveto 8 12 lineto
    4 -4 moveto 10 8 lineto stroke
  }
>> matrix makepattern def
/pixm5 <<
  Tile8x8
  /PaintProc {
    pop 2 6 1 0 360 arc fill 6 2 1 0 360 arc fill
  }
>> matrix makepattern def
/pixm6 <<
  Tile8x8
  /PaintProc {
    pop
    2 2 0.5 0 360 arc fill 2 6 0.5 0 360 arc fill
    6 2 0.5 0 360 arc fill 6 6 0.5 0 360 arc fill
  }
>> matrix makepattern def
/pixm7 <<
  Tile8x8
  /PaintProc {
    pop 0.5 setlinewidth 8 -2 moveto -4 4 lineto 12 0 moveto -4 8 lineto
    12 4 moveto 0 10 lineto stroke
  }
>> matrix makepattern def
/pixm8 <<
  Tile8x8
  /PaintProc {
    pop 0.5 setlinewidth 0 -2 moveto 12 4 lineto -4 0 moveto 12 8 lineto
    -4 4 moveto 8 10 lineto stroke
  }
>> matrix makepattern def\n";
  dd.postscript := Some oc;
  Buffer.clear dd.text_buf;
  expose ();
  draw_ps_text oc (Buffer.contents dd.text_buf);
  fprintf oc "showpage\n";
  dd.postscript := None;
};
