(* $Id: dameid.ml,v 1.746 2011-11-03 13:17:07 deraugla Exp $ *)

open Ast;
open Dtasks;
open Printf;

value schedule_and_wrt ts2 = do {
  let ts = ts2.ast in
  let wrto =
    try do {
      Schedule.schedule_all ts2;
      printf "schedulable\n";
      for i = 0 to ts2.schedule_len - 1 do {
        Sarray.set ts2.disp_schedule i (Sarray.get ts2.comp_schedule i);
      };
      let wprt =
        Array.mapi
          (fun itask td -> (Schedule.compute_wrt ts2 itask, td.period))
          ts.tasks_def
      in
      let (delta, wwt, wwp) =
        Array.fold_left
          (fun (delta, wwt, wwp) (wrt, per) ->
             let d = float wrt /. float per in
             if d > delta then (d, wrt, per) else (delta, wwt, wwp))
          (0.0, 0, 0) wprt
      in
      printf "WRT";
      Array.iter (fun (wrt, _) -> printf " %d" wrt) wprt;
      printf "\n";
      Some (delta, wwt, wwp)
    }
    with
    [ Schedule.Schedule_interrupted _ -> do {
        printf "not schedulable\n";
        None
      } ]
    in
  flush stdout;
  wrto
};

value delta_other =
  fun
  [ [(ts_2, wrto2); (ts_1, wrto1) :: tsl_rest] ->
      let ts1 = ts_1.ast in
      let ts2 = ts_2.ast in
      let comparable =
        loop 0 where rec loop i =
          if i = Array.length ts1.tasks_def && i = Array.length ts2.tasks_def
          then True
          else if
            i = Array.length ts1.tasks_def || i = Array.length ts2.tasks_def
          then False
          else if ts1.tasks_def.(i).period = ts2.tasks_def.(i).period then
            loop (i + 1)
          else False
      in
      if comparable then do {
        let g =
          loop 0.0 (Array.length ts1.tasks_def - 1)
          where rec loop g i =
            if i < 0 then g /. float (Array.length ts1.tasks_def)
            else
              let wrt1 = Schedule.compute_wrt ts_1 i in
              let wrt2 = Schedule.compute_wrt ts_2 i in
              let g = g +. float (wrt2 - wrt1) /. float wrt2 in
              loop g (i - 1)
        in
        printf "gain %.4f\n" g;
        match (wrto1, wrto2) with
        [ (Some (d1, wwt1, wwp1), Some (d2, wwt2, wwp2)) -> do {
             printf "DELTA";
             printf " %.4f (%d/%d)" d2 wwt2 wwp2;
             printf " %.4f (%d/%d)" d1 wwt1 wwp1;
             printf "\n";
             printf "working in sync = %.15f\n"
               (((float wwp2 -. float wwt2) /. float wwp2) /.
                ((float wwp1 -. float wwt1) /. float wwp1));
          }
        | _ -> () ];
        List.iter
          (fun (ts, wrto) ->
             match wrto with
             [ Some (d, wwt, wwp) -> do {
                 printf "OTHER";
                 printf " %.4f (%d/%d)" d wwt wwp;
                 printf "\n";
               }
             | None -> () ])
          tsl_rest;
        printf "\n";
        flush stdout;
      }
      else ()
  | [] | [_] -> () ]
;

value just_schedule fnames algo_opt = do {
  let fname = List.hd fnames in
  let ic =
    if fname = "-" then stdin
    else open_in fname
  in
  let dfl =
    let strm = Stream.of_channel ic in
    match
      try Left (Sparse.parse_and_check_dameid strm) with
      [ Ploc.Exc loc e -> Right (loc, e) ]
    with
    [ Left dl -> List.map snd dl
    | Right (loc, e) -> do {
        if fname = "-" then do {
          let line = Ploc.line_nb loc in
          let bp = Ploc.first_pos loc in
          let ep = Ploc.last_pos loc in
          eprintf "Line %d, characters %d-%d:\n" line bp ep;
        }
        else do {
          let (fname, line, bp, ep) = Ploc.from_file fname loc in
          eprintf "File \"%s\", line %d, characters %d-%d:\n" fname line bp
            ep;
        };
        match e with
        [ Stream.Error s -> eprintf "Parse error: %s\n" s
        | Plexing.Error s -> eprintf "Lexing error: %s\n" s
        | _ -> () ];
        flush stderr;
        failwith "error in dameid input"
      } ]
  in
  let tsl =
    List.map
      (fun df -> do {
         let df =
           match algo_opt with
           [ Some algo -> {(df) with Sparse.d_default_algo = Some algo}
           | None -> df ]
         in
         let ts2 = Scenario.make_tasks df in
         let wrto = schedule_and_wrt ts2 in
         (ts2, wrto)
       })
      dfl
  in
  delta_other tsl;
  if fname <> "-" then close_in ic else ();
};

type mode = [ M_batch | (*M_gtk |*) M_rt ];

value dameid mode fnames algo_opt fail_if_err = do {
  match mode with
  [ M_batch -> just_schedule fnames algo_opt
(*
  | M_gtk -> Graph_gtk.start fnames algo_opt
*)
  | M_rt -> Graph_rt.start fnames algo_opt fail_if_err ]
};

value rev_fnames = ref [];
value fail_if_err = ref False;
value arg_algo = ref None;
value arg_mode = ref M_rt;

value set_algo arg_algo s = do {
  arg_algo.val :=
    match String.lowercase s with
    [ "dm" -> Some AL_deadline_monotonic
    | "rm" -> Some AL_rate_monotonic
    | "edf" -> Some AL_edf
    | _ -> raise (Arg.Bad (sprintf "Bad algorithm: \"%s\"" s)) ];
};

value spec_list =
  Arg.align
    [("-lang", Arg.String I18n.set_lang, "<lang> Set language");
     ("-a", Arg.String (set_algo arg_algo),
      "<algo> Set algorithm (dm, rm, edf)");
     ("-b", Arg.Unit (fun () -> arg_mode.val := M_batch), " Batch mode");
(*
     ("-gtk", Arg.Unit (fun () -> arg_mode.val := M_gtk),
      " Drawing with GTK");
*)
     ("-e", Arg.Set fail_if_err, " Abort program if error (debugging)");
     ("-t", Arg.Set Schedule.trace, " Trace (debugging)");
     ("-", Arg.Unit (fun () -> rev_fnames.val := ["-" :: rev_fnames.val]),
      " Read standard input")]
;
value anon_fun s = rev_fnames.val := [s :: rev_fnames.val];

value main () = do {
  Arg.parse spec_list anon_fun "Usage: dameid";
  dameid arg_mode.val (List.rev rev_fnames.val) arg_algo.val fail_if_err.val
};

main ();
