(* $Id: i18n.ml,v 1.2 2008-03-01 13:23:30 deraugla Exp $ *)

(*
type pure_string = format unit unit unit;
*)
type pure_string = string;
(**)

value lexicon_file =
  ref (Filename.concat (Filename.dirname Sys.argv.(0)) "dameid.lexicon")
;

value lexicon = ref [];
value lexicon_mtime = ref 0.0;
value missing_trace = ref False;

value default_lang =
  try Sys.getenv "LC_ALL" with
  [ Not_found ->
      try Sys.getenv "LC_MESSAGES" with
      [ Not_found -> try Sys.getenv "LANG" with [ Not_found -> "en" ] ] ]
;
value lang = ref default_lang;

value missing str = if missing_trace.val then "{" ^ str ^ "}" else str;

value add_lexicon word transl =
  lexicon.val := [(word, transl) :: lexicon.val]
;

value cut_trail_dot s =
  let len = String.length s in
  if (len >= 3 && s.[len-2] = ' ' || len = 1) && s.[len-1] = '.' then
    String.sub s 0 (len - 1)
  else s
;

value rec interp_slashes s =
  try
    let i = String.index s '\\' in
    let len = String.length s in
    if i + 1 < len then
      match s.[i+1] with
      [ 'n' ->
          String.sub s 0 i ^ "\n" ^
          interp_slashes (String.sub s (i + 2) (len - i - 2))
      | _ -> s ]
    else s
  with
  [ Not_found -> s ]
;

value input_lexicon_line ic =
  loop (input_line ic) where rec loop line =
    let len = String.length line in
    if len > 0 && line.[len-1] = '\\' then
      let new_line =
        let s = input_line ic in
        remove_initial_blanks 0 where rec remove_initial_blanks i =
          if i = String.length s then ""
          else
            match s.[i] with
            [ ' ' | '\t' -> remove_initial_blanks (i + 1)
            | _ -> String.sub s i (String.length s - i) ]
      in
      loop (String.sub line 0 (len - 1) ^ new_line)
    else line
;

value read_lexicon () = do {
  let ic = open_in lexicon_file.val in
  try
    while True do {
      let s =
        let s = input_lexicon_line ic in interp_slashes (cut_trail_dot s)
      in
      let len = String.length s in
      if s = "highlight_missing" then missing_trace.val := True
      else if len >= 4 && String.sub s 0 4 = "    " then
        let s = String.sub s 4 (len - 4) in
        if lang.val <> "" then
          let rec loop default =
            let t =
              try Some (input_lexicon_line ic) with [ End_of_file -> None ]
            in
            let ti =
              match t with
              [ Some t ->
                  try Some (t, String.index t ':') with [ Not_found -> None ]
              | None -> None ]
            in
            match ti with
            [ Some (t, i) ->
                let line_lang = String.sub t 0 i in
                if line_lang = lang.val ||
                   String.length lang.val > String.length line_lang &&
                   String.sub lang.val 0 (String.length line_lang) =
                     line_lang ||
                   s.[0] = '~' && default = "" && line_lang = "en"
                then
                  let t =
                    if i + 2 < String.length t then
                      String.sub t (i + 2) (String.length t - i - 2)
                    else ""
                  in
                  let t = interp_slashes (cut_trail_dot t) in
                  if line_lang = lang.val then add_lexicon s (Some t)
                  else loop t
                else loop default
            | None ->
                add_lexicon s (if default = "" then None else Some default) ]
          in
          loop ""
        else add_lexicon s (Some s)
      else ()
    }
  with
  [ End_of_file -> () ];
  close_in ic
};

value gen_transl str = do {
  if Sys.file_exists lexicon_file.val then
    if lexicon_mtime.val = 0. then do {
      lexicon.val := [];
      lexicon_mtime.val := 1.;
      read_lexicon ()
    }
    else ()
  else ();
  try Some (List.assoc str lexicon.val) with [ Not_found -> None ]
};

value check_format ini_fmt (r : string) =
  let s : string = Obj.magic (ini_fmt : format 'a 'b 'c) in
  loop 0 0 where rec loop i j =
    if i < String.length s - 1 && j < String.length r - 1 then
      match (s.[i], s.[i+1], r.[j], r.[j+1]) with
      [ ('%', x, '%', y) -> if x = y then loop (i + 2) (j + 2) else None
      | ('%', _, _, _) -> loop i (j + 1)
      | (_, _, '%', _) -> loop (i + 1) j
      | _ -> loop (i + 1) (j + 1) ]
    else if i < String.length s - 1 then
      if s.[i] = '%' then None else loop (i + 1) j
    else if j < String.length r - 1 then
      if r.[j] = '%' then None else loop i (j + 1)
    else Some (Obj.magic r : format 'a 'b 'c)
;

value valid_format ini_fmt r =
  match check_format ini_fmt r with
  [ Some fmt -> fmt
  | None -> Obj.magic (missing (Obj.magic ini_fmt)) ]
;

value transl str =
  let str : string = Obj.magic (str : pure_string) in
  match gen_transl str with
  [ Some (Some t) -> t
  | Some None -> missing str
  | None -> if lang.val = "" then str else missing str ]
;

value contain_utf_8_chars s =
  loop 0 where rec loop i =
    if i = String.length s then False
    else if Char.code s.[i] > 127 then True
    else loop (i + 1)
;

value n8transl (s : pure_string) =
  let t = transl s in
  if contain_utf_8_chars t then (Obj.magic s : string) else t
;

value ftransl (fmt : format 'a 'b 'c) =
  let sfmt : string = Obj.magic fmt in
  match gen_transl sfmt with
  [ Some (Some t) -> valid_format fmt t
  | Some None -> Obj.magic missing sfmt
  | None ->
      if lang.val = "" then fmt
      else (Obj.magic (missing sfmt) : format 'a 'b 'c) ]
;

value ptransl str =
  let str : string = Obj.magic (str : pure_string) in
  match gen_transl str with
  [ Some x -> if x = Some str then None else x
  | None -> None ]
;

value highlight_missing v = missing_trace.val := v;

value set_lang new_lang =
  if lang.val <> new_lang then do {
    lang.val := new_lang;
    lexicon_mtime.val := 0.0
  }
  else ()
;

(* *)

value erase str i j =
  String.sub str 0 i ^ String.sub str j (String.length str - j)
;

value eval_set str =
  loop [] str 0 where rec loop set str i =
    if i + 3 < String.length str then
      if str.[i] = '@' && str.[i+1] = '(' && str.[i+3] <> '?' &&
         str.[i+3] <> '-'
      then
        if str.[i+2] = '&' && str.[i+3] = ')' && i + 4 < String.length str
        then
          loop set (erase str i (i + 5)) i
        else
          let (set, j) =
            loop set (i + 2) where rec loop set i =
              if i < String.length str then
                if str.[i] <> ')' then loop [str.[i] :: set] (i + 1)
                else (set, i + 1)
              else (set, i)
          in
          loop set (erase str i j) i
      else loop set str (i + 1)
    else (set, str)
;

value rec apply_expr set str i =
  if i + 1 < String.length str && str.[i+1] = '?' then
    if List.mem str.[i] set then
      let str = erase str i (i + 2) in
      let (str, i) = apply_expr set str i in
      if i < String.length str && str.[i] = ':' then
        let (str, j) = apply_expr set str (i + 1) in (erase str i j, i)
      else (str, i)
    else
      let (str, j) = apply_expr set str (i + 2) in
      let str = erase str i j in
      if i < String.length str && str.[i] = ':' then
        let str = erase str i (i + 1) in apply_expr set str i
      else (str, i)
  else if i < String.length str && (str.[i] = ':' || str.[i] = ')') then
    (str, i)
  else apply_expr set str (i + 1)
;

value eval_app set str =
  loop str 0 where rec loop str i =
    if i + 3 < String.length str then
      if str.[i] = '@' && str.[i+1] = '(' && str.[i+3] <> '-' then
        let str = erase str i (i + 2) in
        let (str, i) = apply_expr set str i in
        if i < String.length str then
          if str.[i] = ')' then loop (erase str i (i + 1)) i else loop str i
        else str
      else loop str (i + 1)
    else str
;

value rec eval_shift s =
  let t = String.make (String.length s) '#' in
  loop False 0 0 where rec loop changed i j =
    if i + 4 < String.length s && s.[i] = '@' && s.[i+1] = '(' &&
       s.[i+3] = '-'
    then
      let nleft = Char.code s.[i+2] - Char.code '0' in
      let to_the_end = s.[i+4] = '-' in
      let k = if to_the_end then i + 5 else i + 4 in
      if k < String.length s && s.[k] = ')' then
        let l =
          loop nleft (i - 1) where rec loop nleft l =
            if l > 0 then
              if s.[l] = ' ' then
                if nleft <= 1 then l + 1 else loop (nleft - 1) (l - 1)
              else loop nleft (l - 1)
            else 0
        in
        let len = i - l in
        let j = j - len in
        let k = k + 1 in
        let i = if k < String.length s && s.[k] = ' ' then k + 1 else k in
        let (i, j) =
          if to_the_end then
            let rec loop i j =
              if i < String.length s then do {
                t.[j] := s.[i];
                loop (i + 1) (j + 1)
              }
              else do {
                let j =
                  if j > 0 && t.[j-1] <> ' ' then do { t.[j] := ' '; j + 1 }
                  else j
                in
                String.blit s l t j len;
                (i, j + len)
              }
            in
            loop i j
          else
            let rec loop i j =
              if i < String.length s then
                if s.[i] = ' ' then do {
                  t.[j] := ' ';
                  String.blit s l t (j + 1) len;
                  (i, j + 1 + len)
                }
                else do { t.[j] := s.[i]; loop (i + 1) (j + 1) }
              else if k < String.length s && s.[k] = ' ' then do {
                t.[j] := ' ';
                String.blit s l t (j + 1) len;
                (i, j + 1 + len)
              }
              else do { String.blit s l t j len; (i, j + len) }
            in
            loop i j
        in
        loop True i j
      else do { t.[j] := s.[i]; loop changed (i + 1) (j + 1) }
    else if i < String.length s then do {
      t.[j] := s.[i];
      loop changed (i + 1) (j + 1)
    }
    else if changed then eval_shift (String.sub t 0 j)
    else String.sub t 0 j
;

value ctransl strl =
  let str = String.concat " " strl in
  let (set, str) = eval_set str in
  let str = eval_app set str in eval_shift str
;
