(* $Id: openfile.ml,v 1.30 2011-03-15 09:42:23 deraugla Exp $ *)

open Printf;

type choice 'a 'b = [ Left of 'a | Right of 'b ];

value save_dir_excursion f = do {
  let ini_wd = Sys.getcwd () in
  let r = try f () with e -> do { Sys.chdir ini_wd; raise e } in
  Sys.chdir ini_wd;
  r
};

value dir_files_nb_rows a =
  loop 0 0 where rec loop max_row i =
    if i = Array.length a then max_row
    else
      let (row, _, _) = a.(i) in
      loop (max max_row row) (i + 1)
;

value dir_files dir ncols = do {
  let files = try Sys.readdir dir with [ Sys_error _ -> [| |] ] in
  let files =
    Array.fold_left
      (fun files fname ->
         if fname <> "" && fname.[0] = '.' then files
         else [fname :: files])
      [] files
  in
  let files =
    save_dir_excursion
      (fun () -> do {
         Sys.chdir dir;
         List.rev_map
           (fun fname ->
              if try Sys.is_directory fname with [ Sys_error _ -> False ]
              then
                fname ^ "/"
              else fname)
           files
       })
  in
  let files = ["../" :: files] in
  let nb_files = List.length files in
  let files = List.sort compare files in
  let len = ref 0 in
  List.iter (fun file -> len.val := max len.val (String.length file)) files;
  len.val := len.val + 2;
  let nbycol = max 1 ((ncols - 1) / len.val) in
  let df =
    loop 0 0 0 [] files where rec loop row col n rev_df =
      fun
      [ [fname :: files] ->
          let row = if n = 0 then row + 1 else row in
          let col = if n = 0 then 0 else col + len.val in
          let n = (n + 1) mod nbycol in
          loop row col n [(row, col, fname) :: rev_df] files
      | [] ->
          Array.of_list (List.rev rev_df) ]
  in
  (df, nb_files)
};

value read_all_dameids fname ic = do {
  let strm = Stream.of_channel ic in
  match
    try Left (Sparse.parse_and_check_dameid strm) with
    [ Ploc.Exc loc e -> Right (loc, e) ]
  with
  [ Left dl -> Some (List.map snd dl)
  | Right (loc, e) -> do {
      if fname = "-" then do {
        let line = Ploc.line_nb loc in
        let bp = Ploc.first_pos loc in
        let ep = Ploc.last_pos loc in
        eprintf "Line %d, characters %d-%d:\n" line bp ep;
      }
      else do {
        let (fname, line, bp, ep) = Ploc.from_file fname loc in
        eprintf "File \"%s\", line %d, characters %d-%d:\n" fname line bp ep;
      };
      match e with
      [ Stream.Error s -> eprintf "Parse error: %s\n" s
      | Plexing.Error s -> eprintf "Lexing error: %s\n" (String.escaped s)
      | _ -> () ];
      flush stderr;
      None
    } ]
};

value get_dameid_files fnames algo_opt = do {
  let (rev_dfnl, rev_bad) =
    List.fold_left
      (fun (rev_dfnl, rev_bad) fname -> do {
         let ic = if fname = "-" then stdin else open_in fname in
         let r =
           match read_all_dameids fname ic with
           [ Some dl ->
              let (rev_dfnl, _) =
                let dl_len = List.length dl in
                List.fold_left
                  (fun (dl, i) d ->
                     let sname =
                       if dl_len = 1 then fname
                       else sprintf "%s [%d]" fname i
                     in
                     let d =
                       match algo_opt with
                       [ Some algo ->
                           {(d) with Sparse.d_default_algo = Some algo}
                       | None -> d ]
                     in
                     ([(d, sname) :: dl], i + 1))
                  (rev_dfnl, 1) dl
              in
              (rev_dfnl, rev_bad)
           | None ->
              (rev_dfnl, [fname :: rev_bad]) ]
         in
         if fname = "-" then () else close_in ic;
         r
       })
      ([], []) fnames
  in
  let dfnl = List.rev rev_dfnl in
  let bad = List.rev rev_bad in
  List.iter
    (fun fname -> do {
       printf "cannot read dameid file \"%s\"\n" fname;
       flush stdout;
     })
     bad;
  dfnl
};

open Glob_glob;

value input_dameid_scenarios tg fname make_scenario = do {
  match try Some (open_in fname) with [ Sys_error _ -> None ] with
  [ Some ic -> do {
      let dlo = read_all_dameids fname ic in
      close_in ic;
      try
        match dlo with
        [ Some dl -> do {
            let dl_len = List.length dl in
            let (rev_scs, _) =
              List.fold_left
                (fun (scs, i) df ->
                   let ts2 = Scenario.make_tasks df in
                   let sname =
                     if dl_len = 1 then fname else sprintf "%s [%d]" fname i
                   in
                   let sc = make_scenario (ts2, sname) in
                   ([sc :: scs], i + 1))
                ([], 1) dl
            in
            let scs = List.rev rev_scs in
            let len = Array.length tg.scenario in
            let isc = if len = 0 then 0 else tg.ind_of_isc.(len-1) + 1 in
            tg.scenario := Array.append tg.scenario (Array.of_list scs);
            tg.ind_of_isc :=
              Array.append tg.ind_of_isc
                (Array.init dl_len (fun i -> isc + i));
            Some len
          }
        | None -> None ]
      with e -> do {
        tg.g_fun.g_error e;
        None
      }
    }
  | None -> None ]
};
