(* $Id: scenario.ml,v 1.69 2011-03-15 09:42:23 deraugla Exp $ *)

open Printf;

open Ast;
open Glob_glob;
open Dtasks;
open Sparse;

(* arbitrary value; if lcm is too big, time or memory can overflow *)
value max_lcm = 500_000_000;

value set_task_order ts2 = do {
  let ts = ts2.ast in
  let tc = ts.constraints in
  let task_order =
    let list =
      let list = Sutil.rho (Array.length ts.tasks_def) in
      match (tc.global_precedence, tc.graph_precedence) with
      [ (True, _) | (_, [_ :: _]) ->
          List.sort (fun it1 it2 -> compare_priorities ts it1 it2) list
      | _ ->
          list ]
    in
    Array.of_list list
  in
  for i = 0 to Array.length ts.tasks_def - 1 do {
    ts2.task_order.(i) := task_order.(i);
  };
};

value make_tasks d = do {
  let ts = Sparse.make_ts d in
  let tdl = Array.to_list ts.tasks_def in
  let ntasks = List.length tdl in
  let task_order = Array.of_list (Sutil.rho ntasks) in
  let (min_rel_time, max_rel_time) =
    List.fold_left
      (fun (min_rel_time, max_rel_time) td ->
         (min min_rel_time td.release, max max_rel_time td.release))
      (max_int, min_int) tdl
  in
  let rmax = max_rel_time - min_rel_time in
  let lcm = Sutil.lcm_of_list (List.map (fun td -> td.period) tdl) in
(*
  if lcm > max_lcm then failwith (sprintf "LCM too big: %d" lcm) else ();
*)
  try do {
    let schedule_len = if rmax = 0 then lcm else rmax + 2 * lcm in
    let disp_schedule = Sarray.make schedule_len None in
    let comp_schedule = Sarray.make schedule_len None in
    let latency_constr =
      List.map (fun l -> {l_constr = l; l_status = LS_unchecked})
        d.d_latency_constr
    in
    let ts2 =
      {ast = ts; task_order = task_order;
       user_graph_precedence = d.d_graph_prec;
       strict_periodicity = d.d_strict_per;
       fully_non_preemptive = d.d_fully_non_pree;
       latency_constr = latency_constr;

       schedule_len = schedule_len;
       disp_schedule = disp_schedule; comp_schedule = comp_schedule;
       rmax = rmax; lcm = lcm; min_rel_time = min_rel_time;

       schedule_status = SS_running; repetition = None;
       curr_ord_ind = 0; curr_inst = 0; curr_step = 0;
       do_refresh = False}
    in
    set_task_order ts2;
    ts2
  }
  with
  [ Out_of_memory ->
      failwith (sprintf "LCM too big: %d (out of memory)" lcm) ]
};

value make_scenario_gen (ts2, fname) graph = do {
  let ts = ts2.ast in
  let tc = ts.constraints in
  let initial_order = Array.init (Array.length ts.tasks_def) (fun i -> i) in
  let show_sub =
    Array.init (Array.length ts.tasks_def)
      (fun itask -> ts.tasks_def.(itask).sub_duration <> [])
  in
  let iparam_of_itask = Array.init (Array.length ts.tasks_def) (fun i -> i) in
  {ts = ts2;
   initial_order = initial_order;
   new_task = Array.copy ts.tasks_def;
   new_preemption_cost = tc.preemption_cost;
   new_graph_precedence = ts2.user_graph_precedence;
   new_latency_constr = ts2.latency_constr;
   new_user_priority = [];
   iparam_of_itask = iparam_of_itask; show_sub = show_sub;
   latency_disp_cnt = 0; expose_sched_cnt = 0; expose_dameid_cnt = 0;
   coloured = True; text_nb_lines = 0; fname = fname;
   graph = graph}
};
