(* $Id: schedule.ml,v 1.256 2011-07-13 14:39:43 deraugla Exp $ *)

open Ast;
open Printf;
open Sparse;
open Dtasks;

exception Schedule_interrupted of schedule_status;

value trace = ref False;

(* Scheduling algorithm *)

value recompute_init ts2 = do {
  let ts = ts2.ast in
  let (min_rel_time, max_rel_time) =
    List.fold_left
      (fun (min_rel_time, max_rel_time) td ->
        let rel = td.release in
        (min min_rel_time rel, max max_rel_time rel))
      (max_int, min_int) (Array.to_list ts.tasks_def)
  in
  let rmax = max_rel_time - min_rel_time in
  ts2.rmax := rmax;
  ts2.min_rel_time := min_rel_time;
  let len = if rmax = 0 then ts2.lcm else rmax + 2 * ts2.lcm in
  ts2.disp_schedule := Sarray.make len None;
  ts2.comp_schedule := Sarray.make len None;
  ts2.schedule_len := len;
  List.iter (fun l -> l.l_status := LS_unchecked) ts2.latency_constr;
};

value make_beg_end_arrays ts2 =
  let ts = ts2.ast in
  let lim = if ts2.rmax = 0 then 0 else ts2.rmax + ts2.lcm in
  let trans_a =
    if lim = 0 then Array.make (Array.length ts.tasks_def) []
    else
      Array.mapi
       (fun itask td ->
          loop [] 0 where rec loop rev_list j =
            if j >= lim then List.rev rev_list
            else
              match Sarray.get ts2.comp_schedule j with
              [ Some u when u.u_task = itask && not u.u_preempt ->
                  loop1 (td.duration - 1) (j + 1) where rec loop1 rest k =
                    if rest = 0 then loop [(j, k) :: rev_list] k
                    else
                      let rest =
                        match Sarray.get ts2.comp_schedule k with
                        [ Some u when u.u_task = itask && not u.u_preempt ->
                            rest - 1
                        | Some _ | None -> rest ]
                      in
                      loop1 rest (k + 1)
              | Some _ | None -> loop rev_list (j + 1) ])
       ts.tasks_def
  in
  let perm_a =
    Array.mapi
      (fun itask td ->
         let start_j =
           if lim = 0 then 0
           else
             let j =
               (ts2.rmax - (td.release - ts2.min_rel_time)) mod td.period
             in
             if j = 0 then ts2.rmax + ts2.lcm
             else ts2.rmax + 2 * ts2.lcm - j
         in
         loop [] start_j where rec loop list j =
           match Sarray.get ts2.comp_schedule j with
           [ Some u when u.u_task = itask && not u.u_preempt ->
               loop1 (td.duration - 1) (j + 1)
               where rec loop1 rest k =
                 if rest = 0 then
                   let list = [(j, k) :: list] in
                   let j = if k = ts2.schedule_len then lim else k in
                   if j = start_j then List.sort compare list
                   else loop list j
                 else
                   let k = if k = ts2.schedule_len then lim else k in
                   let rest =
                     match Sarray.get ts2.comp_schedule k with
                     [ Some u when u.u_task = itask && not u.u_preempt ->
                         rest - 1
                     | Some _ | None -> rest ]
                   in
                   loop1 rest (k + 1)
           | Some _ | None ->
               let j = if j + 1 = ts2.schedule_len then lim else j + 1 in
               if j = start_j then List.sort compare list
               else loop list j ])
      ts.tasks_def
  in
  (trans_a, perm_a)
;

value trace_begin_end ts2 (trans_a, perm_a) = do {
  let ts = ts2.ast in
  for itask = 0 to Array.length ts.tasks_def - 1 do {
    eprintf "task %d\n" (itask + 1);
    eprintf "- transitional:";
    List.iter
      (fun (b, e) ->
         eprintf " (%d, %d)" (b + ts2.min_rel_time) (e + ts2.min_rel_time))
      trans_a.(itask);
    eprintf "\n";
    eprintf "- permanent:";
    List.iter
      (fun (b, e) ->
         eprintf " (%d, %d)" (b + ts2.min_rel_time) (e + ts2.min_rel_time))
      perm_a.(itask);
    eprintf "\n";
    flush stderr;
  };
};

value test_latencies ts2 =
  let ts = ts2.ast in
  if ts2.latency_constr <> [] then do {
    let (trans_a, perm_a) = make_beg_end_arrays ts2 in
    let lim =
      if ts2.rmax = 0 then 0
      else
        match ts2.repetition with
         [ Some (i, _) -> i
         | None -> ts2.rmax + ts2.lcm ]
    in
    List.iter
      (fun l ->
         let lc = l.l_constr in
         let itask1 = lc.l_task1 in
         let itask2 = lc.l_task2 in
         let inst1 = lc.l_inst1 in
         let inst2 = lc.l_inst2 in
         if itask1 >= 0 && itask1 < Array.length ts.tasks_def &&
            itask2 >= 0 && itask2 < Array.length ts.tasks_def &&
            inst1 >= 1 && inst2 >= 1
         then do {
           let ((b1, e1), d1) =
             try (List.nth trans_a.(itask1) (inst1 - 1), 0) with
             [ Failure _ | Invalid_argument _ ->
                 let start_n = List.length trans_a.(itask1) in
                 loop 0 (inst1 - start_n - 1) perm_a.(itask1)
                 where rec loop d n =
                   fun
                   [ [(b, e) :: bel] ->
                       if n = 0 then ((b, e), d) else loop d (n - 1) bel
                   | [] ->
                       loop (d + ts2.lcm) n perm_a.(itask1) ] ]
           in
           let ((b2, e2), d2) =
             try (List.nth trans_a.(itask2) (inst2 - 1), 0) with
             [ Failure _ | Invalid_argument _ ->
                 let start_n = List.length trans_a.(itask2) in
                 loop 0 (inst2 - start_n - 1) perm_a.(itask2)
                 where rec loop d n =
                   fun
                   [ [(b, e) :: bel] ->
                       if n = 0 then ((b, e), d) else loop d (n - 1) bel
                   | [] ->
                       loop (d + ts2.lcm) n perm_a.(itask2) ] ]
           in
           let b1 = b1 + d1 in
           let e1 = e1 + d1 in
           let b2 = b2 + d2 in
           let e2 = e2 + d2 in
           let e1 = if b1 < e1 then e1 else e1 + ts2.lcm in
           if e1 <= b2 || b2 >= lim || e2 >= lim then do {
             let d2 =
               if e1 <= b2 then 0
               else (e1 - b2 + ts2.lcm - 1) / ts2.lcm * ts2.lcm
             in
             let b2 = b2 + d2 in
             let e2 = e2 + d2 in
             let e2 = if b2 < e2 then e2 else e2 + ts2.lcm in
             let lat = e2 - b1 in
             ts2.do_refresh := True;
             l.l_status := LS_values (b1, e1) (b2, e2) lat
           }
           else l.l_status := LS_invalid
         }
         else l.l_status := LS_out_of_bounds)
      ts2.latency_constr;
    loop ts2.latency_constr where rec loop =
      fun
      [ [l :: ll] ->
           let lc = l.l_constr in
           match l.l_status with
           [ LS_values b e lat -> if lat > lc.l_max then False else loop ll
           | LS_invalid | LS_out_of_bounds | LS_unchecked -> loop ll ]
      | [] -> True ];
  }
  else True
;

value check_latencies ts =
  if test_latencies ts then ()
  else
    let msg = I18n.transl "latency constraint failure" in
    raise (Schedule_interrupted (SS_schedulable_but msg))
;

type schedule_instance_param =
  { si_init_part : bool;
    si_initial_start_j : int;
    si_start_j : int;
    si_j : int;
    si_is_first_unit : bool;
    si_check_task_strict_per : bool;
    si_dur_rest : list Ast.sub_duration;
    si_current_duration : int;
    si_normal_rest : int;
    si_preempt_rest : int;
    si_next_j : int -> int }
;

type schedule_instance_result =
  [ SI_instance_done of schedule_instance_param
  | SI_task_ended
  | SI_task_error of string ]
;

value pay_preempted_preemption = False;

(* schedule_instance *)

value rec schedule_instance ts2 td itask instance si =
  let ts = ts2.ast in
  let tc = ts.constraints in
  if si.si_init_part && si.si_j >= ts2.schedule_len then
    if ts2.rmax = 0 && si.si_normal_rest > 0 then
      SI_task_error (I18n.transl "system load > 1")
    else
      SI_task_ended
  else if si.si_normal_rest = 0 then
    match si.si_dur_rest with
    [ [] ->
        let new_start_j =
          let i = si.si_start_j + td.period in
          if not si.si_init_part && i >= ts2.schedule_len then i - ts2.lcm
          else i
        in
        if si.si_init_part && new_start_j >= ts2.schedule_len ||
           not si.si_init_part && new_start_j = si.si_initial_start_j
        then
          SI_task_ended
        else
          let si =
            {(si) with
             si_start_j = new_start_j; si_j = new_start_j;
             si_is_first_unit = True;
             si_check_task_strict_per = td.strict_per;
             si_dur_rest = td.sub_duration;
             si_current_duration = first_duration td;
             si_normal_rest = first_duration td;
             si_preempt_rest = 0}
          in
          SI_instance_done si
    | [sd :: dur_rest] ->
        let j =
          let i = max si.si_j (si.si_start_j + sd.sub_rel) in
          if not si.si_init_part && i >= ts2.schedule_len then i - ts2.lcm
          else i
        in
        let si =
          {(si) with
           si_j = j; si_is_first_unit = True;
           si_check_task_strict_per = sd.sub_str_per;
           si_dur_rest = dur_rest;
           si_current_duration = sd.sub_dur;
           si_normal_rest = sd.sub_dur;
           si_preempt_rest = 0}
        in
        schedule_instance ts2 td itask instance si ]
  else
    let initial_j = si.si_j in
    let j =
      let non_preemptive = ts2.fully_non_preemptive || td.non_preemptive in
      if non_preemptive && si.si_is_first_unit then
        loop si.si_current_duration si.si_j si.si_j
        where rec loop rest good_start j =
          if rest = 0 then Left good_start
          else if si.si_init_part && j = ts2.schedule_len then
            Right (I18n.transl "non preemptive task failure")
          else
            let j1 = si.si_next_j j in
            let (rest, good_start) =
              if Sarray.get ts2.comp_schedule j <> None then
                (si.si_current_duration, j1)
              else
                (rest - 1, good_start)
            in
            if not si.si_init_part && j1 = si.si_initial_start_j && rest > 0
            then
              Right (I18n.transl "non preemptive task failure")
            else
              loop rest good_start j1
      else
        loop si.si_j where rec loop j =
          if si.si_init_part && j = ts2.schedule_len then Left j
          else if Sarray.get ts2.comp_schedule j = None then Left j
          else
            let j1 = si.si_next_j j in
            if not si.si_init_part && j1 = si.si_initial_start_j then
              Right (I18n.transl "system load > 1")
            else loop j1
    in
    match j with
    [ Left j ->
        let _ =
          assert
            (j = ts2.schedule_len || Sarray.get ts2.comp_schedule j = None)
        in
        if (ts2.strict_periodicity || si.si_check_task_strict_per) &&
           si.si_is_first_unit && j <> initial_j
        then
          SI_task_error (I18n.transl "strict periodicity failed")
        else if
          si.si_init_part && ts2.rmax > 0 && j >= ts2.rmax + ts2.lcm
        then
          if si.si_normal_rest > 0 && j - si.si_start_j >= td.deadline then
            let msg =
              if td.period = td.deadline then I18n.transl "period exceeded"
              else I18n.transl "deadline exceeded"
            in
            SI_task_error msg
          else
            let new_initial_start_j =
              let j =
                (ts2.rmax - (td.release - ts2.min_rel_time)) mod td.period
              in
              if j = 0 then ts2.rmax + ts2.lcm
              else ts2.rmax + 2 * ts2.lcm - j
            in
            let next_j j =
              if j + 1 < ts2.schedule_len then j + 1 else j + 1 - ts2.lcm
            in
            let si =
              {si_init_part = False;
               si_initial_start_j = new_initial_start_j;
               si_start_j = new_initial_start_j;
               si_j = new_initial_start_j;
               si_is_first_unit = True;
               si_check_task_strict_per = td.strict_per;
               si_dur_rest = td.sub_duration;
               si_current_duration = first_duration td;
               si_normal_rest = first_duration td;
               si_preempt_rest = 0;
               si_next_j = next_j}
            in
            schedule_instance ts2 td itask instance si
        else
          let (j, normal_rest, preempt_rest) =
            if si.si_init_part && j = ts2.schedule_len then
              (j, si.si_normal_rest, si.si_preempt_rest)
            else do {
              let preemption_cost =
                if td.pree_cost > 0 then td.pree_cost
                else tc.preemption_cost
              in
              let preemption =
                not si.si_is_first_unit && j <> initial_j &&
                preemption_cost > 0
              in
              let u =
                {u_task = itask; u_inst = instance;
                 u_preempt = preemption || si.si_preempt_rest > 0}
              in
              let (normal_rest, preempt_rest) =
                if preemption then
                  let preempt_rest =
                    if pay_preempted_preemption then
                      si.si_preempt_rest - 1 + preemption_cost
                    else
                      - 1 + preemption_cost
                  in
                  (si.si_normal_rest, preempt_rest)
                else if si.si_preempt_rest > 0 then
                  (si.si_normal_rest, si.si_preempt_rest - 1)
                else
                  (si.si_normal_rest - 1, 0)
              in
              Sarray.set ts2.comp_schedule j (Some u);
              let j = si.si_next_j j in
              (j, normal_rest, preempt_rest)
            }
          in
          let response_time =
            if si.si_init_part || j > si.si_start_j then j - si.si_start_j
            else j - si.si_start_j + ts2.lcm
          in
          if response_time > td.deadline then
            let msg =
              if td.deadline = td.period then I18n.transl "period exceeded"
              else I18n.transl "deadline exceeded"
            in
            SI_task_error msg
          else
            let si =
              {(si) with
               si_j = j; si_is_first_unit = False;
               si_check_task_strict_per = False;
               si_normal_rest = normal_rest;
               si_preempt_rest = preempt_rest}
            in
            schedule_instance ts2 td itask instance si
    | Right msg -> SI_task_error msg ]
;

value schedule_instance_init ts2 td =
  let initial_start_j = td.release - ts2.min_rel_time in
  let next_j j = j + 1 in
  {si_init_part = True; si_initial_start_j = initial_start_j;
   si_start_j = initial_start_j; si_j = initial_start_j;
   si_is_first_unit = True;
   si_check_task_strict_per = td.strict_per;
   si_dur_rest = td.sub_duration;
   si_current_duration = first_duration td;
   si_normal_rest = first_duration td;
   si_preempt_rest = 0;
   si_next_j = next_j}
;

value schedule_task_before_curr_pos ts2 iorder =
  let ts = ts2.ast in
  let itask = ts2.task_order.(iorder) in
  let td = ts.tasks_def.(itask) in
  let si = schedule_instance_init ts2 td in
  let is_last_task = iorder = ts2.curr_ord_ind in
  loop 0 si where rec loop inst si =
    if is_last_task && inst > ts2.curr_inst then inst
    else
      match schedule_instance ts2 td itask inst si with
      [ SI_instance_done si -> loop (inst + 1) si
      | SI_task_ended -> inst
      | SI_task_error msg ->
          raise (Schedule_interrupted (SS_task_error itask msg)) ]
;

value schedule_task ts2 iorder =
  let ts = ts2.ast in
  let itask = ts2.task_order.(iorder) in
  let td = ts.tasks_def.(itask) in
  let si = schedule_instance_init ts2 td in
  loop 0 si where rec loop inst si =
    match schedule_instance ts2 td itask inst si with
    [ SI_instance_done si -> do {
        ts2.curr_step := ts2.curr_step + 1;
        loop (inst + 1) si
      }
    | SI_task_ended -> ts2.curr_step := ts2.curr_step + 1
    | SI_task_error msg ->
        raise (Schedule_interrupted (SS_task_error itask msg)) ]
;

value update_release ts2 scheduling_tasks_list iorder =
  let ts = ts2.ast in
  let tc = ts.constraints in
  let itask = ts2.task_order.(iorder) in
  let is_last_task = iorder = ts2.curr_ord_ind in
  let to_restart =
    let preceding_tasks = graph_preceding_tasks tc.graph_precedence itask in
    if (tc.global_precedence || preceding_tasks <> []) && is_last_task then
      let rel =
        loop 0 ts2.min_rel_time where rec loop jorder max_rel =
          if jorder = iorder then max_rel
          else
            let jtask = ts2.task_order.(jorder) in
            if tc.global_precedence || List.mem jtask preceding_tasks then
              let td = ts.tasks_def.(jtask) in
              let rel =
                loop (td.release - ts2.min_rel_time) td.duration
                where rec loop j rest =
                  match Sarray.get ts2.comp_schedule j with
                  [ Some u ->
                      if u.u_task = jtask && not u.u_preempt then
                        if rest = 1 then j + 1 + ts2.min_rel_time
                        else loop (j + 1) (rest - 1)
                      else loop (j + 1) rest
                  | None -> loop (j + 1) rest ]
              in
              loop (jorder + 1) (max max_rel rel)
            else loop (jorder + 1) max_rel
      in
      let td = ts.tasks_def.(itask) in
      if td.release < rel then do {
        ts.tasks_def.(itask) := {(td) with release = rel};
        True
      }
      else False
    else False
  in
  if to_restart then True
  else
    let to_restart =
      (* if strict periodicity and precedence constraint, try to shift
         the release to the first available time unit to give a chance
         to that strict periodicity *)
      let preceding_tasks = graph_preceding_tasks tc.graph_precedence itask in
      let td = ts.tasks_def.(itask) in
      if (tc.global_precedence || preceding_tasks <> []) && is_last_task &&
         (ts2.strict_periodicity || td.strict_per)
      then
        let j =
          loop (td.release - ts2.min_rel_time) where rec loop j =
            if j >= ts2.schedule_len then failwith "next_instance"
            else if Sarray.get ts2.comp_schedule j = None then j
            else loop (j + 1)
        in
        if td.release <> j + ts2.min_rel_time then do {
          ts.tasks_def.(itask) := {(td) with release = j + ts2.min_rel_time};
          True
        }
        else False
      else False
    in
    to_restart
;

value compute_repetitive_part ts2 =
  if ts2.rmax = 0 then None
  else
    let period =
      match ts2.schedule_status with
      [ SS_several_lcm n -> n * ts2.lcm
      | SS_running | SS_incomplete_scheduling _ _ | SS_task_error _ _
      | SS_schedulable_but _ -> ts2.lcm ]
    in
    loop (ts2.schedule_len - period - 1) where rec loop i =
      if i <= 0 then Some (0, period)
      else
        match
          (Sarray.get ts2.comp_schedule i,
           Sarray.get ts2.comp_schedule (i+period))
        with
        [ (Some u1, Some u2) ->
            if u1.u_task = u2.u_task && u1.u_preempt = u2.u_preempt then
              loop (i - 1)
            else Some (i + 1, period)
        | (None, None) -> loop (i - 1)
        | _ -> Some (i + 1, period) ]
;

value rec schedule_next_instance_aux ts2 =
  let ts = ts2.ast in
  if ts2.curr_ord_ind = Array.length ts.tasks_def then ()
  else if ts2.schedule_status <> SS_running then ()
  else do {
    for i = 0 to ts2.schedule_len - 1 do {
      Sarray.set ts2.comp_schedule i None
    };
    let scheduling_tasks_list =
      List.sort
        (fun i1 i2 ->
           compare_priorities ts ts2.task_order.(i1) ts2.task_order.(i2))
        (Sutil.rho (ts2.curr_ord_ind + 1))
    in
    loop scheduling_tasks_list where rec loop =
      fun
      [ [iorder :: rest] ->
          if update_release ts2 scheduling_tasks_list iorder then do {
            recompute_init ts2;
            ts2.do_refresh := True;
            schedule_next_instance_aux ts2
          }
          else do {
            let inst = schedule_task_before_curr_pos ts2 iorder in
            let is_last_task = iorder = ts2.curr_ord_ind in
            if is_last_task then do {
              if inst > ts2.curr_inst then ts2.curr_inst := ts2.curr_inst + 1
              else do {
                ts2.curr_ord_ind := ts2.curr_ord_ind + 1;
                ts2.curr_inst := 0;
              };
              ts2.curr_step := ts2.curr_step + 1;
            }
            else ();
            loop rest
          }
      | [] -> do {
          List.iter (fun l -> l.l_status := LS_unchecked) ts2.latency_constr;
          if ts2.curr_ord_ind = Array.length ts.tasks_def then do {
            match compute_repetitive_part ts2 with
            [ Some r -> do {
                ts2.repetition := Some r;
                ts2.do_refresh := True;
              }
            | None -> () ];
            check_latencies ts2
          }
          else ()
        } ]
  }
;

type task_info_for_edf =
  { instance : int;
    started : bool;
    preemption : int;
    duration_fifo : list duration_info;
    next_deadline : int;
    next_release : int;
    preceding_tasks : list int }
and duration_info =
  { di_remaining : int;
    di_non_preemptive : bool;
    di_strict_periodic : bool }
;

value compute_number_of_lcm ts2 =
  loop 2 ts2.schedule_len where rec loop i j =
    if j - 1 - i * ts2.lcm < 0 then None
    else if j = ts2.schedule_len - (* i * *) ts2.lcm then Some i
    else
      match
        (Sarray.get ts2.comp_schedule (j - 1),
         Sarray.get ts2.comp_schedule (j - 1 - i * ts2.lcm))
      with
      [ (Some u1, Some u2) ->
          if u1.u_task = u2.u_task && u1.u_preempt = u2.u_preempt then
            loop i (j - 1)
          else
            loop (i + 1) ts2.schedule_len
      | (None, None) -> loop i (j - 1)
      | _ -> loop (i + 1) ts2.schedule_len ]
;

value check_edf_last_lcm_looping ts2 =
  let ts = ts2.ast in
  for itask = 0 to Array.length ts.tasks_def - 1 do {
    let td = ts.tasks_def.(itask) in
    let ifirst =
      loop (td.release - ts2.min_rel_time) where rec loop i =
        if i >= ts2.schedule_len - ts2.lcm then i
        else loop (i + td.period)
    in
    loop ifirst ifirst td.duration where rec loop istart i rest =
      if i = istart + td.period then
        if rest <> 0 then do {
          let n_lcm = compute_number_of_lcm ts2 in
          let ss =
            match n_lcm with
            [ Some i ->
                SS_several_lcm i
            | None ->
                let msg = I18n.transl "last LCM not looping" in
                SS_incomplete_scheduling itask msg ]
          in
          raise (Schedule_interrupted ss)
        }
        else if i = ifirst + ts2.lcm then ()
        else loop i i td.duration
      else
        let j = if i >= ts2.schedule_len then i - ts2.lcm else i in
        match Sarray.get ts2.comp_schedule j with
        [ Some u ->
            let rest =
              if u.u_task = itask && not u.u_preempt then rest - 1
              else rest
            in
            loop istart (i + 1) rest
        | None ->
            loop istart (i + 1) rest ];
  }
;

value next_dur_dea_rel ts2 itask i =
  let ts = ts2.ast in
  let td = ts.tasks_def.(itask) in
  let pos_in_period = (ts2.min_rel_time + i - td.release) mod td.period in
  let (durinf, sdl) =
    if pos_in_period mod td.period = 0 then
      let durinf =
        {di_remaining = first_duration td;
         di_non_preemptive = td.non_preemptive;
         di_strict_periodic = td.strict_per}
      in
      let sdl = td.sub_duration in
      (durinf, sdl)
    else
      loop td.sub_duration where rec loop =
        fun
        [ [sd :: sdl] ->
            if sd.sub_rel = pos_in_period then
              let durinf =
                {di_remaining = sd.sub_dur;
                 di_non_preemptive = sd.sub_non_pree;
                 di_strict_periodic = sd.sub_str_per}
              in
              (durinf, sdl)
            else
              loop sdl
        | [] ->
            failwith
              (sprintf "next_dur_dea_rel: i %d t%d pos_in_period %d" i
                 (itask+1) pos_in_period) ]
  in
  let rel =
    match sdl with
    [ [sd :: sdl] -> i - pos_in_period + sd.sub_rel
    | [] -> i - pos_in_period + td.period ]
  in
  let dea = i - pos_in_period + td.deadline in
  let d_inst = if pos_in_period = 0 then 1 else 0 in
  (durinf, dea, rel, d_inst)
;

value rec schedule_edf_upto ts current_i = do {
  let tc = ts.ast.constraints in
  for i = 0 to ts.schedule_len - 1 do {
    Sarray.set ts.comp_schedule i None
  };
  ts.curr_step := 0;
  let task_list =
    loop 0 [] where rec loop itask rev_list =
      if itask = Array.length ts.ast.tasks_def then List.rev rev_list
      else
        let td = ts.ast.tasks_def.(itask) in
        let preceding_tasks =
          graph_preceding_tasks tc.graph_precedence itask
        in
        let ti =
          if td.release = ts.min_rel_time then
            let (durinf, dea, rel, _) = next_dur_dea_rel ts itask 0 in
            {instance = 1; started = False; preemption = 0;
             duration_fifo = [durinf]; next_deadline = dea;
             next_release = rel; preceding_tasks = preceding_tasks}
          else
            {instance = 0; started = False; preemption = 0;
             duration_fifo = []; next_deadline = 0;
             next_release = td.release - ts.min_rel_time;
             preceding_tasks = preceding_tasks}
        in
        loop (itask + 1) [(itask, ti) :: rev_list]
  in
  loop 0 task_list where rec loop i task_list = do {
    if i = ts.schedule_len then do {
      (* hack to specify end *)
      ts.curr_ord_ind := Array.length ts.ast.tasks_def;
      check_edf_last_lcm_looping ts;
      check_latencies ts;
    }
    else do {
      let rev_cand_tl =
        loop [] task_list where rec loop rev_cand_tl =
          fun
          [ [(itask, ti) :: task_list] ->
              if ti.preceding_tasks <> [] then loop rev_cand_tl task_list
              else
                match ti.duration_fifo with
                [ [di :: dil] ->
                    let non_preemptive =
                      ts.fully_non_preemptive || di.di_non_preemptive
                    in
                    if ti.started && non_preemptive then [itask]
                    else loop [itask :: rev_cand_tl] task_list
                | [] ->
                    loop rev_cand_tl task_list ]
          | [] -> rev_cand_tl ]
      in
      match List.rev rev_cand_tl with
      [ [] -> do {
          let i =
            List.fold_left (fun i (itask, ti) -> min i ti.next_release)
              max_int task_list
          in
          let task_list =
            List.map
              (fun (itask, ti) ->
                 let ti =
                   if i = ti.next_release then
                     let (durinf, dea, rel, d_inst) =
                       next_dur_dea_rel ts itask i
                     in
                     let durfif = ti.duration_fifo @ [durinf] in
                     {instance = ti.instance + d_inst; started = False;
                      preemption = 0; duration_fifo = durfif;
                      next_deadline = dea; next_release = rel;
                      preceding_tasks = ti.preceding_tasks}
                   else
                     {instance = ti.instance; started = False;
                      preemption = 0; duration_fifo = ti.duration_fifo;
                      next_deadline = 0; next_release = ti.next_release;
                      preceding_tasks = ti.preceding_tasks}
                 in
                 (itask, ti))
              task_list
          in
          loop i task_list
        }
      | cand_tl -> do {
          let itask =
            match cand_tl with
            [ [itask] -> itask
            | cand_tl ->
                (* choosing the closest deadline if any *)
                let (closest_dead, rev_cand_tl) =
                  List.fold_left
                    (fun (closest_dead, rev_cand_tl) (itask, ti) ->
                       if ti.duration_fifo <> [] && ti.next_deadline > 0 then
                         if ti.next_deadline < closest_dead then
                           (ti.next_deadline, [itask])
                         else if closest_dead < ti.next_deadline then
                           (closest_dead, rev_cand_tl)
                         else
                           (closest_dead, [itask :: rev_cand_tl])
                       else
                         (closest_dead, rev_cand_tl))
                    (max_int, []) task_list
                in
                match List.rev rev_cand_tl with
                [ [itask] -> itask
                | cand_tl ->
                    (* choosing the one which was already running *)
                    loop (i - 1) where rec loop j =
                      if j < 0 then
                        (* choosing the first candidate *)
                        List.hd cand_tl
                      else
                        match Sarray.get ts.comp_schedule j with
                        [ Some u ->
                            if List.mem u.u_task cand_tl then
                              let ti = List.assoc u.u_task task_list in
                              if ti.started then u.u_task
                              else List.hd cand_tl
                            else
                              List.hd cand_tl
                        | None ->
                            loop (j - 1) ] ] ]
          in
          let ti = List.assoc itask task_list in
          let task_list =
            (* possibly preempting a task *)
            match
              if i > 0 then Sarray.get ts.comp_schedule (i - 1) else None
            with
            [ Some u ->
                let pree_cost =
                  let td_u = ts.ast.tasks_def.(u.u_task) in
                  if td_u.pree_cost > 0 then td_u.pree_cost
                  else tc.preemption_cost
                in
                if u.u_task <> itask && pree_cost > 0 then
                  List.map
                    (fun (itask2, ti) ->
                       let ti =
                         if itask2 = u.u_task then
                           match ti.duration_fifo with
                           [ [di :: dil] ->
                               if ti.started && not di.di_non_preemptive then
                                 {(ti) with preemption = pree_cost}
                               else
                                 ti
                           | [] -> ti ]
                         else
                           ti
                       in
                       (itask2, ti))
                    task_list
                else task_list
            | None -> task_list ]
          in
          let is_first_instance_ending =
            (* for knowing whether this task is ending its first instance
               and has a precedence constraint with other tasks *)
            let ti = List.assoc itask task_list in
            if ti.instance = 1 then
              match ti.duration_fifo with
              [ [di] -> di.di_remaining = 1
              | [] | [_ :: _] -> False ]
            else False
          in
          let (rev_task_list, preemption, must_restart) =
            loop ([], False) task_list
            where rec loop (rev_task_list, preemption) =
              fun
              [ [(itask2, ti) :: ttl] ->
                 let (ti, preemption, must_restart) =
                   if itask2 = itask then
                     if ti.preemption > 0 then
                       ({(ti) with preemption = ti.preemption - 1},
                        True, False)
                     else
                       let (duration_fifo, started) =
                         match ti.duration_fifo with
                         [ [di :: dil] ->
                             let rem = di.di_remaining in
                             if rem = 1 then (dil, False)
                             else
                               let di = {(di) with di_remaining = rem - 1} in
                               ([di :: dil], True)
                         | [] -> assert False ]
                       in
                       ({(ti) with
                        started = started; duration_fifo = duration_fifo},
                        False, False)
                   else if
                     is_first_instance_ending && ti.preceding_tasks <> []
                   then do {
                     let pt = List.filter (\<> itask) ti.preceding_tasks in
                     let td2 = ts.ast.tasks_def.(itask2) in
                     let new_rel = ts.min_rel_time + i + 1 in
                     let must_restart = pt = [] && td2.release < new_rel in
                     if must_restart then
                       let td2 = {(td2) with release = new_rel} in
                       ts.ast.tasks_def.(itask2) := td2
                     else ();
                     ({(ti) with preceding_tasks = pt}, preemption,
                      must_restart)
                   }
                   else (ti, preemption, False)
                 in
                 let rev_task_list = [(itask2, ti) :: rev_task_list] in
                 if must_restart then
                   (rev_task_list, preemption, True)
                 else
                   loop (rev_task_list, preemption) ttl
              | [] ->
                  (rev_task_list, preemption, False) ]
          in
          if must_restart then do {
            let old_len = ts.schedule_len in
            recompute_init ts;
            let new_len = ts.schedule_len in
            ts.do_refresh := True;
            schedule_edf_upto ts (current_i - old_len + new_len)
          }
          else do {
            let u =
              {u_task = itask; u_inst = ti.instance; u_preempt = preemption}
            in
            Sarray.set ts.comp_schedule i (Some u);
            ts.curr_step := ts.curr_step + 1;
            List.iter
              (fun (itask2, ti2) ->
                 match ti2.duration_fifo with
                 [ [di2 :: dil2] ->
                     if (ts.strict_periodicity || di2.di_strict_periodic) &&
                        not ti2.started && itask2 <> itask
                     then
                       let msg = I18n.transl "strict periodicity failed" in
                       raise (Schedule_interrupted (SS_task_error itask2 msg))
                     else ()
                 | [] -> () ])
              task_list;
            if i >= current_i then ()
            else
              let i = i + 1 in
              let task_list =
                List.rev_map
                  (fun (itask, ti) ->
                     if ti.next_deadline = i && ti.duration_fifo <> [] then
                       let msg =
                         if ti.next_deadline = ti.next_release then
                           I18n.transl "period exceeded"
                         else
                           I18n.transl "deadline exceeded"
                       in
                       raise (Schedule_interrupted (SS_task_error itask msg))
                     else
                       let ti =
                         if ti.next_release = i then
                           let (durinf, dea, rel, d_inst) =
                             next_dur_dea_rel ts itask i
                           in
                           let durfif = ti.duration_fifo @ [durinf] in
                           {instance = ti.instance + d_inst; started = False;
                            preemption = ti.preemption;
                            duration_fifo = durfif;
                            next_deadline = dea; next_release = rel;
                            preceding_tasks = ti.preceding_tasks}
                         else
                           ti
                       in
                       (itask, ti))
                  rev_task_list
              in
              loop i task_list
          }
        } ]
    }
  }
};

value next_edf_schedule ts2 = do {
  let current_i =
    loop (ts2.schedule_len - 1) where rec loop i =
      match Sarray.get ts2.disp_schedule i with
      [ Some _ -> i + 1
      | None -> if i = 0 then 0 else loop (i - 1) ]
  in
  schedule_edf_upto ts2 current_i;
(*
  schedule_edf_upto ts2 ts2.curr_step;
  ts2.curr_step := ts2.curr_step + 1;
*)
};

value schedule_next_instance ts2 =
  let ts = ts2.ast in
  if ts.algorithm = AL_edf then next_edf_schedule ts2
  else schedule_next_instance_aux ts2
;

value rec schedule_all_normal ts2 = do {
  let ts = ts2.ast in
  let scheduling_tasks_list =
    List.sort
      (fun i1 i2 ->
         compare_priorities ts ts2.task_order.(i1) ts2.task_order.(i2))
      (Sutil.rho (Array.length ts.tasks_def))
  in
  for i = 0 to ts2.schedule_len - 1 do {
    Sarray.set ts2.comp_schedule i None
  };
  loop scheduling_tasks_list where rec loop =
    fun
    [ [iorder :: rest] -> do {
         ts2.curr_ord_ind := iorder;
         if update_release ts2 scheduling_tasks_list iorder then do {
           recompute_init ts2;
           ts2.curr_step := 0;
           ts2.do_refresh := True;
           schedule_all_normal ts2
         }
         else do {
           schedule_task ts2 iorder;
           loop rest
         }
       }
    | [] -> do {
        List.iter (fun l -> l.l_status := LS_unchecked) ts2.latency_constr;
        ts2.curr_ord_ind := Array.length ts.tasks_def;
        ts2.curr_inst := 0;
        match compute_repetitive_part ts2 with
        [ Some r -> do {
            ts2.repetition := Some r;
            ts2.do_refresh := True;
          }
        | None -> () ];
        check_latencies ts2;
      } ];
};

(* alternative version for Audsley *)
value rec gen_schedule_all2 ts2 = do {
  let ts = ts2.ast in
  let scheduling_tasks_list = Sutil.rho (Array.length ts.tasks_def) in
  for i = 0 to ts2.schedule_len - 1 do {
    Sarray.set ts2.comp_schedule i None
  };
  loop scheduling_tasks_list where rec loop =
    fun
    [ [iorder :: rest] -> do {
         ts2.curr_ord_ind := iorder;
         if update_release ts2 scheduling_tasks_list iorder then do {
           recompute_init ts2;
           ts2.do_refresh := True;
           gen_schedule_all2 ts2
         }
         else
           let itask = ts2.task_order.(iorder) in
           let td = ts.tasks_def.(itask) in
           let si = schedule_instance_init ts2 td in
           loop_task 0 si where rec loop_task inst si =
             match schedule_instance ts2 td itask inst si with
             [ SI_instance_done si -> loop_task (inst + 1) si
             | SI_task_ended -> loop rest
             | SI_task_error msg -> Some itask ]
       }
    | [] -> do {
        List.iter (fun l -> l.l_status := LS_unchecked) ts2.latency_constr;
        ts2.curr_ord_ind := Array.length ts.tasks_def;
        ts2.curr_inst := 0;
        ts2.curr_step := 0;
        match compute_repetitive_part ts2 with
        [ Some r -> do {
            ts2.repetition := Some r;
            ts2.do_refresh := True;
          }
        | None -> () ];
        if test_latencies ts2 then None else
        Some (Array.length ts.tasks_def - 1)
      } ];
};

value schedule_all_edf ts2 =
  let ts = ts2.ast in
  if ts2.curr_ord_ind < Array.length ts.tasks_def then
    schedule_edf_upto ts2 ts2.schedule_len
  else do {
    (* hack to test other lcm *)
    ts2.schedule_status := SS_running;
    ts2.repetition := None;
    ts2.schedule_len := ts2.schedule_len + ts2.lcm;
    ts2.comp_schedule := Sarray.make ts2.schedule_len None;
    ts2.disp_schedule := Sarray.make ts2.schedule_len None;
    schedule_edf_upto ts2 ts2.schedule_len;
  }
;

value schedule_all ts2 =
  let ts = ts2.ast in
  if ts.algorithm = AL_edf then schedule_all_edf ts2
  else schedule_all_normal ts2
;

value goto_begin_schedule ts2 = do {
  let ts = ts2.ast in
  let tc = ts.constraints in
  if tc.global_precedence || tc.graph_precedence <> [] then
    for itask = 0 to Array.length ts.tasks_def - 1 do {
      let td = ts.tasks_def.(itask) in
      let r = if tc.global_precedence then 0 else td.user_release in
      let td = {(td) with release = r} in
      ts.tasks_def.(itask) := td;
    }
  else ();
  Sarray.reinit ts2.disp_schedule None;
  Sarray.reinit ts2.comp_schedule None;
  ts2.schedule_status := SS_running;
  ts2.curr_ord_ind := 0;
  ts2.curr_inst := 0;
  ts2.curr_step := 0;
  ts2.repetition := None;
  recompute_init ts2;
};

value before pt t1 t2 =
  loop pt where rec loop =
    fun
    [ [t :: tl] ->
        if t = t1 then True
        else if t = t2 then False
        else loop tl
    | [] -> invalid_arg "before" ]
;

value compatible_priority ts pt =
  let tc = ts.constraints in
  List.for_all
    (fun tl ->
       loop tl where rec loop =
         fun
         [ [t1 :: ([t2 :: _] as tl)] ->
             if before pt t1 t2 then loop tl else False
         | [_] | [] -> True ])
    tc.graph_precedence
;

value first_permut n =
  loop n [] where rec loop i list =
    if i <= 0 then list else loop (i - 1) [i - 1 :: list]
;
value last_permut n =
  loop 1 [] where rec loop i list =
    if i > n then list else loop (i + 1) [i - 1 :: list]
;

value rec rev_next_permut comp right =
  fun
  [ [x :: rlist] ->
      loop [] right where rec loop rleft =
        fun
        [ [y :: right] ->
            if comp x y then
              List.rev_append rlist [y :: List.rev_append rleft [x :: right]]
            else
              loop [y :: rleft] right
        | [] ->
            rev_next_permut comp (List.rev [x :: rleft]) rlist ]
  | [] -> [] ]
;

value next_permut list = rev_next_permut \< [] (List.rev list);
value prev_permut list = rev_next_permut \> [] (List.rev list);

value rec fact n = if n <= 1 then 1 else n * fact (n - 1);

value next_prev_audsley_schedule ts2 step last_pos is_next =
  let ts = ts2.ast in
  let len = Array.length ts.tasks_def in
  let ntot = fact len in
  let (nlim, next_prev_permut, one) =
    if is_next then (ntot, next_permut, 1) else (0, prev_permut, -1)
  in
  match last_pos with
  [ Some (last_permut, n, nsol) ->
      loop last_permut n where rec loop permut n =
        let prio_list = permut in
        if permut = [] then do {
          step (nlim, ntot);
          False
        }
        else if compatible_priority ts prio_list then do {
          step (n + 1, ntot);
          let t = Array.of_list prio_list in
          let a =
            {ai_pos = P_middle; ai_tab = t; ai_step = n; ai_nsol = nsol}
          in
          let ts = {(ts) with algorithm = AL_audsley a} in
          ts2.ast := ts;
          ts2.task_order := t;
          goto_begin_schedule ts2;
          match gen_schedule_all2 ts2 with
          [ Some itask ->
              let (permut, n) =
                loop [] permut where rec loop r =
                  fun
                  [ [i :: l] ->
                      if i = itask then
                        let (l, n) =
                          loop l n where rec loop l n =
                            let l1 = next_prev_permut l in
                            if l1 = [] then (l, n)
                            else loop l1 (n + one)
                        in
                        (List.rev_append [i :: r] l, n)
                      else
                        loop [i :: r] l
                  | [] -> assert False ]
              in
              loop (next_prev_permut permut) (n + one)
          | None -> True ]
        }
        else
          loop (next_prev_permut permut) (n + one)
  | None -> do {
      step (nlim, ntot);
      False
    } ]
;

value next_audsley_schedule ts2 step auds =
  let ts = ts2.ast in
  let last_pos =
    match auds.ai_pos with
    [ P_begin ->
        let len = Array.length ts.tasks_def in
        Some (first_permut len, 0, 1)
    | P_middle ->
        let permut = next_permut (Array.to_list auds.ai_tab) in
        Some (permut, auds.ai_step + 1, auds.ai_nsol + 1)
    | P_end -> None ]
  in
  next_prev_audsley_schedule ts2 step last_pos True
;

value prev_audsley_schedule ts2 step auds =
  let ts = ts2.ast in
  let last_pos =
    match auds.ai_pos with
    [ P_begin -> None
    | P_middle ->
        let permut = prev_permut (Array.to_list auds.ai_tab) in
        Some (permut, auds.ai_step - 1, auds.ai_nsol - 1)
    | P_end ->
        let len = Array.length ts.tasks_def in
        Some (last_permut len, fact len - 1, auds.ai_nsol) ]
  in
  next_prev_audsley_schedule ts2 step last_pos False
;

value rec schedule_entire_task ts2 =
  let ts = ts2.ast in
  let curr_ord_ind = ts2.curr_ord_ind in
  if curr_ord_ind = Array.length ts.tasks_def then ()
  else do {
    let scheduling_tasks_list =
      List.sort
        (fun i1 i2 ->
           compare_priorities ts ts2.task_order.(i1) ts2.task_order.(i2))
        (Sutil.rho (curr_ord_ind + 1))
    in
    for i = 0 to ts2.schedule_len - 1 do {
      Sarray.set ts2.comp_schedule i None
    };
    loop scheduling_tasks_list where rec loop =
      fun
      [ [iorder :: rest] -> do {
           ts2.curr_ord_ind := iorder;
           if update_release ts2 scheduling_tasks_list iorder then do {
             recompute_init ts2;
             ts2.do_refresh := True;
             schedule_entire_task ts2
           }
           else do {
             schedule_task ts2 iorder;
             loop rest
           }
         }
      | [] -> do {
          ts2.curr_ord_ind := curr_ord_ind + 1;
          ts2.curr_inst := 0;
          ts2.curr_step := 0;
          if ts2.curr_ord_ind = Array.length ts.tasks_def then do {
            List.iter (fun l -> l.l_status := LS_unchecked)
              ts2.latency_constr;
            match compute_repetitive_part ts2 with
            [ Some r -> do {
                ts2.repetition := Some r;
                ts2.do_refresh := True;
              }
            | None -> () ];
            check_latencies ts2;
          }
          else ();
        } ];
    }
;

(* Printing *)

value string_list_of_graph_prec gp =
  List.map
    (List.fold_left
       (fun s tn ->
          let n = "t" ^ string_of_int (tn + 1) in
          if s = "" then n else s ^ "<" ^ n)
       "")
    gp
;

value string_of_graph_prec gp =
  let sl = string_list_of_graph_prec gp in
  String.concat " " sl
;

value string_of_one_latency l =
  let lc = l.l_constr in
  sprintf "t%d(%d)->t%d(%d)<%d" (lc.l_task1 + 1) lc.l_inst1
    (lc.l_task2 + 1) lc.l_inst2 lc.l_max
;

value string_of_latency ll =
  let sl = List.map string_of_one_latency ll in
  String.concat " " sl
;

value string_of_one_user_priority l =
  sprintf "t%d" (l + 1)
;

value string_of_user_priority ll =
  let sl = List.map string_of_one_user_priority ll in
  String.concat "," sl
;

value string_of_priority_array a =
  let ll = do {
    let ll = ref [] in
    for i = 0 to Array.length a - 1 do {
      ll.val := [(a.(i), i) :: ll.val];
    };
    List.map snd (List.sort compare ll.val)
  }
  in
  string_of_user_priority ll
;

open Std_term;

type vt =
  { vt_bold : string;
    vt_normal : string;
    vt_color : int -> int -> int -> string;
    vt_no_color : string;
    vt_erase_line_from_cursor : string }
;

value vt_term =
  {vt_bold = vt_bold; vt_normal = vt_normal;
   vt_color = vt_color; vt_no_color = vt_no_color;
   vt_erase_line_from_cursor = vt_erase_line_from_cursor}
;

value vt_pango =
  {vt_bold = "\\<b>";
   vt_normal = "\\</b>";
   vt_color r g b = sprintf "\\<span foreground=\"#%02x%02x%02x\">" r g b;
   vt_no_color = "\\</span>"; vt_erase_line_from_cursor = ""}
;

value constr_info_list vt ts2 ncols =
  let ts = ts2.ast in
  let tc = ts.constraints in
  let list1 = [] in
  let list1 =
    let txt =
      sprintf "%s%s %s" (I18n.transl "priority") (I18n.transl ":")
        (match ts.algorithm with
         [ AL_deadline_monotonic -> I18n.transl "deadline-monotonic"
         | AL_rate_monotonic -> I18n.transl "rate-monotonic"
         | AL_edf -> I18n.transl "earliest deadline first"
         | AL_audsley a ->
             let s = I18n.transl "Audsley++" in
             if a.ai_pos = P_middle then
               let a = Array.to_list a.ai_tab in
               let sl = List.map string_of_one_user_priority a in
               s ^ " " ^ String.concat "," sl
             else s
         | AL_user a ->
             string_of_priority_array a ])
    in
    [txt :: list1]
  in
  let list1 =
    let has_preemption_costs =
      List.exists (fun td -> td.pree_cost > 0) (Array.to_list ts.tasks_def)
    in
    if has_preemption_costs then list1
    else
      let txt =
        sprintf "%s%s %d" (I18n.transl "preemption cost") (I18n.transl ":")
          tc.preemption_cost
      in
      [txt :: list1]
  in
  let number_of_non_preemptive_tasks =
    List.fold_left (fun n td -> if td.non_preemptive then n + 1 else n)
      0 (Array.to_list ts.tasks_def)
  in
  let graph_precedence = tc.graph_precedence in
  let list1 =
    if tc.global_precedence || graph_precedence <> [] ||
       ts2.strict_periodicity || ts2.fully_non_preemptive ||
       ts2.latency_constr <> [] || number_of_non_preemptive_tasks > 0
    then
      [I18n.transl "system constraints" ^ I18n.transl ":" :: list1]
    else list1
  in
  let list1 =
    if tc.global_precedence then
      ["- " ^ I18n.transl "precedence" :: list1]
    else if graph_precedence <> [] then
      let txt = "- " ^ I18n.transl "precedence" ^ I18n.transl ":" in
      let txtl = string_list_of_graph_prec graph_precedence in
      loop list1 "" txtl where rec loop list1 s1 sl =
        match sl with
        [ [s2 :: sl] ->
            if s1 <> "" && String.length (txt ^ s1 ^ " " ^ s2) > ncols then
              loop [txt ^ s1 :: list1] "" [s2 :: sl]
            else
              loop list1 (s1 ^ " " ^ s2) sl
        | [] ->
            if s1 <> "" then [txt ^ s1 :: list1] else list1 ]
    else list1
  in
  let list2 =
    if ts2.latency_constr <> [] then
      let (sl, slen) =
        List.fold_right
          (fun l (sl, slen) ->
             let txt =
               "- " ^ I18n.transl "latency" ^ I18n.transl ":" ^ " " ^
               string_of_one_latency l
             in
             let stat =
               if ts2.curr_ord_ind = Array.length ts.tasks_def &&
                  match ts2.schedule_status with
                  [ SS_running | SS_schedulable_but _ -> True
                  | _ -> False ]
               then
                 match l.l_status with
                 [ LS_unchecked -> "bug"
                 | LS_invalid | LS_out_of_bounds -> I18n.transl "ignored"
                 | LS_values (beg1, _) (_, end2) lat ->
                     let lc = l.l_constr in
                     let failed = lat > lc.l_max in
                     sprintf ":= %s%d%s"
                       (if failed then vt.vt_color 255 0 0 else "")
                       lat (if failed then " !" ^ vt.vt_no_color else "") ]
               else ""
             in
             ([(txt, stat) :: sl], max slen (String.length txt)))
           ts2.latency_constr ([], 0)
      in
      List.fold_left
        (fun list (txt, stat) ->
           let pad =
             if stat = "" then ""
             else String.make (slen + (*2*)1 - String.length txt) ' '
           in
           [txt ^ pad ^ stat :: list])
        [] sl
    else []
  in
  let list3 = [] in
  let list3 =
    if ts2.strict_periodicity then
      ["- " ^ String.uncapitalize (I18n.transl "Strict Periodicity") :: list3]
    else if
      List.exists (fun td -> td.strict_per) (Array.to_list ts.tasks_def)
    then
      let txt =
        let s = String.uncapitalize (I18n.transl "Strict Periodicity") in
        loop ("- " ^ s ^ I18n.transl ":") 0
        where rec loop txt i =
          if i = Array.length ts.tasks_def then txt
          else if ts.tasks_def.(i).strict_per then
            loop (sprintf "%s t%d" txt (i + 1)) (i + 1)
          else loop txt (i + 1)
      in
      [txt :: list3]
    else list3
  in
  let list3 =
    if ts2.fully_non_preemptive then
      ["- " ^ String.uncapitalize (I18n.transl "Fully Non Preemptive") ::
       list3]
    else if number_of_non_preemptive_tasks > 0 then
      let txt =
        let s =
          if number_of_non_preemptive_tasks = 1 then
            I18n.transl "non preemptive task"
          else
            I18n.transl "non preemptive tasks"
        in
        loop ("- " ^ s ^ I18n.transl ":") 0
        where rec loop txt i =
          if i = Array.length ts.tasks_def then txt
          else if ts.tasks_def.(i).non_preemptive then
            loop (sprintf "%s t%d" txt (i + 1)) (i + 1)
          else loop txt (i + 1)
      in
      [txt :: list3]
    else list3
  in
  (List.rev list1, List.rev list2, List.rev list3)
;

value get_response_values ts2 =
  let ts = ts2.ast in
  let n = ts2.schedule_len in
  loop [] 0 where rec loop rev_response_values itask =
    if itask = Array.length ts.tasks_def then
      List.rev rev_response_values
    else
      let td = ts.tasks_def.(itask) in
      let list =
        let start = td.release - ts2.min_rel_time in
        loop [] start start where rec loop rev_list start_period j =
          if j >= n then List.rev rev_list
          else
            let jmax = j + td.deadline in
            loop_count rev_list j td.duration
            where rec loop_count rev_list j count =
              if count = 0 then
                let rev_list =
                  if j > jmax then rev_list
                  else [(start_period, j - start_period) :: rev_list]
                in
                loop rev_list (start_period + td.period)
                  (max j (start_period + td.period))
              else
                let count =
                  let k =
                    loop j where rec loop j =
                      if j < n then j
                      else loop (j - ts2.lcm)
                  in
                  match Sarray.get ts2.disp_schedule k with
                  [ Some u ->
                      if u.u_task = itask && not u.u_preempt then count - 1
                      else count
                  | None -> count ]
                in
                if j > jmax then
                  loop rev_list (start_period + td.period)
                    (start_period + td.period)
                else loop_count rev_list (j + 1) count
      in
      let rev_response_values =
        if list = [] then rev_response_values
        else [(itask, list) :: rev_response_values]
      in
      loop rev_response_values (itask + 1)
;

(* Saving dameid file *)

value print_dameid oc ts2 = do {
  let ts = ts2.ast in
  let tc = ts.constraints in
  fprintf oc "%d\n" tc.preemption_cost;
  for i = 0 to Array.length ts.tasks_def - 1 do {
    let td = ts.tasks_def.(i) in
    let release = td.user_release in
    fprintf oc "%d" release;
    fprintf oc " %d" td.duration;
    if td.sub_duration = [] then ()
    else do {
      fprintf oc " ";
      List.iter
        (fun sd ->
           fprintf oc "(%d %d%s)" sd.sub_rel sd.sub_dur
             (if sd.sub_non_pree then " np" else ""))
        td.sub_duration;
    };
    fprintf oc " %d" td.deadline;
    if td.period <> td.deadline then fprintf oc " %d" td.period else ();
    if td.non_preemptive then fprintf oc " np" else ();
    if td.pree_cost > 0 then fprintf oc " pc %d" td.pree_cost else ();
    fprintf oc "\n";
  };
  if tc.global_precedence then fprintf oc "prec\n" else ();
  if ts2.user_graph_precedence <> [] then
    fprintf oc "prec: %s\n" (string_of_graph_prec ts2.user_graph_precedence)
  else ();
  if ts2.strict_periodicity then fprintf oc "strp\n" else ();
  if ts2.fully_non_preemptive then fprintf oc "fnp\n" else ();
  if ts2.latency_constr <> [] then
    fprintf oc "lat: %s\n" (string_of_latency ts2.latency_constr)
  else ();
};

value output_task_data fname ts2 = do {
  let oc = open_out fname in
  print_dameid oc ts2;
  close_out oc;
};

value total_duration td = td.duration;

value compute_wrt ts2 itask =
  let ts = ts2.ast in
  let td = ts.tasks_def.(itask) in
  loop_all 0 (td.release - ts2.min_rel_time)
  where rec loop_all wrt start_j =
    loop_inst (total_duration td) start_j where rec loop_inst cnt j =
      if cnt = 0 then
        let pwrt = j - start_j in
        loop_all (max wrt pwrt) (start_j + td.period)
      else if j >= ts2.schedule_len then wrt
      else
        let cnt =
          match Sarray.get ts2.disp_schedule j with
          [ Some u ->
              if u.u_task = itask && not u.u_preempt then cnt - 1
              else cnt
          | None ->
              cnt ]
        in
        loop_inst cnt (j + 1)
;

value compute_classical_utilization_factor ts2 =
  let ts = ts2.ast in
  Array.fold_left (fun cuf td -> cuf + td.duration * ts2.lcm / td.period)
    0 ts.tasks_def
;

value compute_exact_utilization ts2 =
  let j = ts2.schedule_len - ts2.lcm in
  loop 0 j where rec loop eu j =
    if j = ts2.schedule_len then eu
    else
      let eu = if Sarray.get ts2.disp_schedule j = None then eu else eu + 1 in
      loop eu (j + 1)
;

open Glob_glob;

value total_duration td = td.duration;

value gen_redraw_text print vt ncols sc = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let tc = ts.constraints in
  let cuf = compute_classical_utilization_factor ts2 in
  let nl () = do {
    print vt.vt_erase_line_from_cursor;
    print "\r\n"
  }
  in
  print vt.vt_bold;
  sc.text_nb_lines := 0;
  print (I18n.transl "classical utilization factor" ^ I18n.transl ":");
  nl ();
  sc.text_nb_lines := sc.text_nb_lines + 1;
  if cuf > ts2.lcm then print (vt.vt_color 256 0 0)
  else ();
  print
    (sprintf "       %4d/%d = %1.4f" cuf ts2.lcm
       (float cuf /. float ts2.lcm));
  nl ();
  nl ();
  sc.text_nb_lines := sc.text_nb_lines + 1;
  sc.text_nb_lines := sc.text_nb_lines + 1;
  if cuf > ts2.lcm then print vt.vt_no_color else ();
  let ended =
    Array.length ts.tasks_def > 0 &&
    ts2.curr_ord_ind = Array.length ts.tasks_def ||
    ts2.schedule_status <> SS_running
  in
  let ended_and_schedulable =
    ended &&
    match ts2.schedule_status with
    [ SS_running | SS_schedulable_but _ | SS_several_lcm _ -> True
    | SS_task_error _ _ | SS_incomplete_scheduling _ _ -> False ]
  in
  if ended_and_schedulable then do {
    let exact_utilization = compute_exact_utilization ts2 in
    print
      (I18n.transl "exact permanent utilization factor" ^ I18n.transl ":");
    nl ();
    sc.text_nb_lines := sc.text_nb_lines + 1;
    print
      (sprintf "       %4d/%d = %1.4f" exact_utilization ts2.lcm
         (float exact_utilization /. float ts2.lcm));
    nl (); nl ();
    sc.text_nb_lines := sc.text_nb_lines + 1;
    sc.text_nb_lines := sc.text_nb_lines + 1;
    print (I18n.transl "exact permanent preemption cost" ^ I18n.transl ":");
    nl ();
    sc.text_nb_lines := sc.text_nb_lines + 1;
    print
      (sprintf "       %4d/%d = %1.4f" (exact_utilization - cuf)
         ts2.lcm (float (exact_utilization - cuf) /. float ts2.lcm));
    nl (); nl ();
    sc.text_nb_lines := sc.text_nb_lines + 1;
    sc.text_nb_lines := sc.text_nb_lines + 1;
  }
  else ();
  let deadline_exists =
    List.exists (fun td -> td.deadline <> td.period)
      (Array.to_list ts.tasks_def)
  in
  let has_preemption_costs =
    List.exists (fun td -> td.pree_cost > 0) (Array.to_list ts.tasks_def)
  in
  if Array.length ts.tasks_def > 0 then do {
    let shorten txt =
      if has_preemption_costs then
        let m = 6 in
        if utf_8_string_length txt > m then utf_8_string_sub txt 0 m else txt
      else txt
    in
    let task_txt = shorten (I18n.transl "task") in
    print task_txt;
    print " ";
    let rel_txt = shorten (I18n.transl "release") in
    print rel_txt;
    print " ";
    let dur_txt = shorten (I18n.transl "duration") in
    print dur_txt;
    let wrt_txt = "   " ^ shorten (I18n.transl "WRT") in
    if ended_and_schedulable then do {
      print " ";
      print vt.vt_normal;
      print wrt_txt;
      print vt.vt_bold;
    }
    else ();
    let dea_txt = shorten (I18n.transl "deadline") in
    if deadline_exists then do {
      print " ";
      print dea_txt;
    }
    else ();
    print " ";
    let per_txt = shorten (I18n.transl "period") in
    print per_txt;
    let prco_txt = shorten (I18n.transl "preemption") in
    if has_preemption_costs then do {
      print " ";
      print prco_txt;
    }
    else ();
    let align s1 s2 =
      let string_length s = utf_8_string_length s in
      String.make (max 0 (string_length s1 - string_length s2 - 1)) ' ' ^ s2
    in
    nl ();
    sc.text_nb_lines := sc.text_nb_lines + 1;
    for itask = 0 to Array.length ts.tasks_def - 1 do {
      let td = ts.tasks_def.(itask) in
      let txt = sprintf "t%d" (itask + 1) in
      print (align task_txt txt);
      print "  ";
      let txt = string_of_int td.release in
      print (align rel_txt txt);
      print "  ";
      let txt = string_of_int (total_duration td) in
      print (align dur_txt txt);
      if ended_and_schedulable then do {
        print "  ";
        let wrt = compute_wrt ts2 itask in
        let txt = string_of_int wrt in
        print vt.vt_normal;
        print (align wrt_txt txt);
        print vt.vt_bold;
      }
      else ();
      if deadline_exists then do {
        print "  ";
        let txt = string_of_int td.deadline in
        print (align dea_txt txt);
      }
      else ();
      print "  ";
      let txt = string_of_int td.period in
      print (align per_txt txt);
      if has_preemption_costs then do {
        let pree_cost = max td.pree_cost tc.preemption_cost in
        print "  ";
        let txt = string_of_int pree_cost in
        print (align prco_txt txt);
      }
      else ();
      nl ();
      sc.text_nb_lines := sc.text_nb_lines + 1;
    };
    nl ();
  }
  else ();
  sc.text_nb_lines := sc.text_nb_lines + 1;
  let (list1, list2, list3) = constr_info_list vt ts2 ncols in
  List.iter
    (fun txt -> do {
       print txt;
       nl ();
       sc.text_nb_lines := sc.text_nb_lines + 1;
     })
    list1;
  let nb_good_latencies =
    List.fold_left
      (fun n lat ->
         match lat.l_status with
         [ LS_values _ _ _ -> n + 1
         | _ -> n ])
      0 ts2.latency_constr
  in
  loop
    (if nb_good_latencies <= 1 then -1
     else sc.latency_disp_cnt mod nb_good_latencies)
    ts2.latency_constr list2
  where rec loop cnt lat_list =
    fun
    [ [txt :: txt_list] -> do {
        let (txt, cnt, lat_list) =
          match lat_list with
          [ [{l_status = LS_values _ _ _} :: lat_list] -> do {
              let txt =
                if cnt = 0 then "*" ^ String.sub txt 1 (String.length txt - 1)
                else txt
              in
              (txt, cnt - 1, lat_list)
            }
          | [_ :: lat_list] -> (txt, cnt, lat_list)
          | [] -> (txt, cnt, []) ]
        in
        print txt;
        nl ();
        sc.text_nb_lines := sc.text_nb_lines + 1;
        loop cnt lat_list txt_list
      }
    | [] -> () ];
  List.iter
    (fun txt -> do {
       print txt;
       nl ();
       sc.text_nb_lines := sc.text_nb_lines + 1;
     })
    list3;
  if ended || cuf > ts2.lcm then do {
    nl ();
    sc.text_nb_lines := sc.text_nb_lines + 1;
    match ts2.schedule_status with
    [ SS_schedulable_but msg -> do {
        print (I18n.transl "SCHEDULABLE BUT...");
        nl ();
        print msg;
        sc.text_nb_lines := sc.text_nb_lines + 2;
      }
    | SS_task_error itask s -> do {
        print (I18n.transl "NOT SCHEDULABLE");
        nl ();
        print (sprintf "%s (t%d)" s (itask + 1));
        sc.text_nb_lines := sc.text_nb_lines + 1;
      }
    | SS_incomplete_scheduling itask s -> do {
        print (I18n.transl "SCHEDULING NOT COMPLETED");
        nl ();
        print (sprintf "%s (t%d)" s (itask + 1));
        sc.text_nb_lines := sc.text_nb_lines + 2;
      }
    | SS_several_lcm n -> do {
        print (I18n.transl "SCHEDULABLE BUT...");
        nl ();
        print
          (I18n.transl
             (sprintf (I18n.ftransl "permanent phase needs %d lcm") n));
        sc.text_nb_lines := sc.text_nb_lines + 2;
      }
    | SS_running ->
        if cuf > ts2.lcm then do {
          print (I18n.transl "NOT SCHEDULABLE");
          nl ();
          print (I18n.transl "system load > 1");
          sc.text_nb_lines := sc.text_nb_lines + 2;
        }
        else
          print (I18n.transl "SCHEDULABLE") ];
    nl ();
    sc.text_nb_lines := sc.text_nb_lines + 1;
  }
  else ();
  print vt.vt_normal;
};
